// chessboard.js v1.0.0
// https://github.com/oakmac/chessboardjs/
//
// Copyright (c) 2019, Chris Oakman
// Released under the MIT license
// https://github.com/oakmac/chessboardjs/blob/master/LICENSE.md

// start anonymous scope
;(function () {
  'use strict'

  var $ = window['jQuery']

  // ---------------------------------------------------------------------------
  // Constants
  // ---------------------------------------------------------------------------

  var COLUMNS = 'abcdefgh'.split('')
  var DEFAULT_DRAG_THROTTLE_RATE = 20
  var ELLIPSIS = '…'
  var MINIMUM_JQUERY_VERSION = '1.8.3'
  var RUN_ASSERTS = false
  var START_FEN = 'rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR'
  var START_POSITION = fenToObj(START_FEN)

  // default animation speeds
  var DEFAULT_APPEAR_SPEED = 200
  var DEFAULT_MOVE_SPEED = 200
  var DEFAULT_SNAPBACK_SPEED = 60
  var DEFAULT_SNAP_SPEED = 30
  var DEFAULT_TRASH_SPEED = 100

  // use unique class names to prevent clashing with anything else on the page
  // and simplify selectors
  // NOTE: these should never change
  var CSS = {}
  CSS['alpha'] = 'alpha-d2270'
  CSS['black'] = 'black-3c85d'
  CSS['board'] = 'board-b72b1'
  CSS['chessboard'] = 'chessboard-63f37'
  CSS['clearfix'] = 'clearfix-7da63'
  CSS['highlight1'] = 'highlight1-32417'
  CSS['highlight2'] = 'highlight2-9c5d2'
  CSS['notation'] = 'notation-322f9'
  CSS['numeric'] = 'numeric-fc462'
  CSS['piece'] = 'piece-417db'
  CSS['row'] = 'row-5277c'
  CSS['sparePieces'] = 'spare-pieces-7492f'
  CSS['sparePiecesBottom'] = 'spare-pieces-bottom-ae20f'
  CSS['sparePiecesTop'] = 'spare-pieces-top-4028b'
  CSS['square'] = 'square-55d63'
  CSS['white'] = 'white-1e1d7'
  CSS['overlay'] = 'overlay-4fc5e'
  CSS['overlayArrow'] = 'overlay-arrow-672a1'
  CSS['overlayGroup'] = 'overlay-group-9d6ed'
  CSS['overlayCircle'] = 'overlay-circle-t69xy'

  // ---------------------------------------------------------------------------
  // Misc Util Functions
  // ---------------------------------------------------------------------------
  function throttle (f, interval, scope) {
    var timeout = 0
    var shouldFire = false
    var args = []

    var handleTimeout = function () {
      timeout = 0
      if (shouldFire) {
        shouldFire = false
        fire()
      }
    }

    var fire = function () {
      timeout = window.setTimeout(handleTimeout, interval)
      f.apply(scope, args)
    }

    return function (_args) {
      args = arguments
      if (!timeout) {
        fire()
      } else {
        shouldFire = true
      }
    }
  }

  // function debounce (f, interval, scope) {
  //   var timeout = 0
  //   return function (_args) {
  //     window.clearTimeout(timeout)
  //     var args = arguments
  //     timeout = window.setTimeout(function () {
  //       f.apply(scope, args)
  //     }, interval)
  //   }
  // }

  function uuid () {
    return 'xxxx-xxxx-xxxx-xxxx-xxxx-xxxx-xxxx-xxxx'.replace(/x/g, function (c) {
      var r = (Math.random() * 16) | 0
      return r.toString(16)
    })
  }

  function deepCopy (thing) {
    return JSON.parse(JSON.stringify(thing))
  }

  function parseSemVer (version) {
    var tmp = version.split('.')
    return {
      major: parseInt(tmp[0], 10),
      minor: parseInt(tmp[1], 10),
      patch: parseInt(tmp[2], 10)
    }
  }

  // returns true if version is >= minimum
  function validSemanticVersion (version, minimum) {
    version = parseSemVer(version)
    minimum = parseSemVer(minimum)

    var versionNum = (version.major * 100000 * 100000) +
                     (version.minor * 100000) +
                     version.patch
    var minimumNum = (minimum.major * 100000 * 100000) +
                     (minimum.minor * 100000) +
                     minimum.patch

    return versionNum >= minimumNum
  }

  function interpolateTemplate (str, obj) {
    for (var key in obj) {
      if (!obj.hasOwnProperty(key)) continue
      var keyTemplateStr = '{' + key + '}'
      var value = obj[key]
      while (str.indexOf(keyTemplateStr) !== -1) {
        str = str.replace(keyTemplateStr, value)
      }
    }
    return str
  }

  if (RUN_ASSERTS) {
    console.assert(interpolateTemplate('abc', {a: 'x'}) === 'abc')
    console.assert(interpolateTemplate('{a}bc', {}) === '{a}bc')
    console.assert(interpolateTemplate('{a}bc', {p: 'q'}) === '{a}bc')
    console.assert(interpolateTemplate('{a}bc', {a: 'x'}) === 'xbc')
    console.assert(interpolateTemplate('{a}bc{a}bc', {a: 'x'}) === 'xbcxbc')
    console.assert(interpolateTemplate('{a}{a}{b}', {a: 'x', b: 'y'}) === 'xxy')
  }

  // ---------------------------------------------------------------------------
  // Predicates
  // ---------------------------------------------------------------------------

  function isString (s) {
    return typeof s === 'string'
  }

  function isFunction (f) {
    return typeof f === 'function'
  }

  function isInteger (n) {
    return typeof n === 'number' &&
           isFinite(n) &&
           Math.floor(n) === n
  }

  function validAnimationSpeed (speed) {
    if (speed === 'fast' || speed === 'slow') return true
    if (!isInteger(speed)) return false
    return speed >= 0
  }

  function validThrottleRate (rate) {
    return isInteger(rate) &&
           rate >= 1
  }

  function validMove (move) {
    // move should be a string
    if (!isString(move)) return false

    // move should be in the form of 'e2-e4', 'f6-d5'
    var squares = move.split('-')
    if (squares.length !== 2) return false

    return validSquare(squares[0]) && validSquare(squares[1])
  }

  function validSquare (square) {
    return isString(square) && square.search(/^[a-h][1-8]$/) !== -1
  }

  if (RUN_ASSERTS) {
    console.assert(validSquare('a1'))
    console.assert(validSquare('e2'))
    console.assert(!validSquare('D2'))
    console.assert(!validSquare('g9'))
    console.assert(!validSquare('a'))
    console.assert(!validSquare(true))
    console.assert(!validSquare(null))
    console.assert(!validSquare({}))
  }

  function validPieceCode (code) {
    return isString(code) && code.search(/^[bw][KQRNBP]$/) !== -1
  }

  if (RUN_ASSERTS) {
    console.assert(validPieceCode('bP'))
    console.assert(validPieceCode('bK'))
    console.assert(validPieceCode('wK'))
    console.assert(validPieceCode('wR'))
    console.assert(!validPieceCode('WR'))
    console.assert(!validPieceCode('Wr'))
    console.assert(!validPieceCode('a'))
    console.assert(!validPieceCode(true))
    console.assert(!validPieceCode(null))
    console.assert(!validPieceCode({}))
  }

  function validFen (fen) {
    if (!isString(fen)) return false

    // cut off any move, castling, etc info from the end
    // we're only interested in position information
    fen = fen.replace(/ .+$/, '')

    // expand the empty square numbers to just 1s
    fen = expandFenEmptySquares(fen)

    // FEN should be 8 sections separated by slashes
    var chunks = fen.split('/')
    if (chunks.length !== 8) return false

    // check each section
    for (var i = 0; i < 8; i++) {
      if (chunks[i].length !== 8 ||
          chunks[i].search(/[^kqrnbpKQRNBP1]/) !== -1) {
        return false
      }
    }

    return true
  }

  if (RUN_ASSERTS) {
    console.assert(validFen(START_FEN))
    console.assert(validFen('8/8/8/8/8/8/8/8'))
    console.assert(validFen('r1bqkbnr/pppp1ppp/2n5/1B2p3/4P3/5N2/PPPP1PPP/RNBQK2R'))
    console.assert(validFen('3r3r/1p4pp/2nb1k2/pP3p2/8/PB2PN2/p4PPP/R4RK1 b - - 0 1'))
    console.assert(!validFen('3r3z/1p4pp/2nb1k2/pP3p2/8/PB2PN2/p4PPP/R4RK1 b - - 0 1'))
    console.assert(!validFen('anbqkbnr/8/8/8/8/8/PPPPPPPP/8'))
    console.assert(!validFen('rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/'))
    console.assert(!validFen('rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBN'))
    console.assert(!validFen('888888/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR'))
    console.assert(!validFen('888888/pppppppp/74/8/8/8/PPPPPPPP/RNBQKBNR'))
    console.assert(!validFen({}))
  }

  function validPositionObject (pos) {
    if (!$.isPlainObject(pos)) return false

    for (var i in pos) {
      if (!pos.hasOwnProperty(i)) continue

      if (!validSquare(i) || !validPieceCode(pos[i])) {
        return false
      }
    }

    return true
  }

  if (RUN_ASSERTS) {
    console.assert(validPositionObject(START_POSITION))
    console.assert(validPositionObject({}))
    console.assert(validPositionObject({e2: 'wP'}))
    console.assert(validPositionObject({e2: 'wP', d2: 'wP'}))
    console.assert(!validPositionObject({e2: 'BP'}))
    console.assert(!validPositionObject({y2: 'wP'}))
    console.assert(!validPositionObject(null))
    console.assert(!validPositionObject('start'))
    console.assert(!validPositionObject(START_FEN))
  }

  function isTouchDevice () {
    return 'ontouchstart' in document.documentElement
  }

  function validJQueryVersion () {
    return typeof window.$ &&
           $.fn &&
           $.fn.jquery &&
           validSemanticVersion($.fn.jquery, MINIMUM_JQUERY_VERSION)
  }

  // ---------------------------------------------------------------------------
  // Chess Util Functions
  // ---------------------------------------------------------------------------

  // convert FEN piece code to bP, wK, etc
  function fenToPieceCode (piece) {
    // black piece
    if (piece.toLowerCase() === piece) {
      return 'b' + piece.toUpperCase()
    }

    // white piece
    return 'w' + piece.toUpperCase()
  }

  // convert bP, wK, etc code to FEN structure
  function pieceCodeToFen (piece) {
    var pieceCodeLetters = piece.split('')

    // white piece
    if (pieceCodeLetters[0] === 'w') {
      return pieceCodeLetters[1].toUpperCase()
    }

    // black piece
    return pieceCodeLetters[1].toLowerCase()
  }

  // convert FEN string to position object
  // returns false if the FEN string is invalid
  function fenToObj (fen) {
    if (!validFen(fen)) return false

    // cut off any move, castling, etc info from the end
    // we're only interested in position information
    fen = fen.replace(/ .+$/, '')

    var rows = fen.split('/')
    var position = {}

    var currentRow = 8
    for (var i = 0; i < 8; i++) {
      var row = rows[i].split('')
      var colIdx = 0

      // loop through each character in the FEN section
      for (var j = 0; j < row.length; j++) {
        // number / empty squares
        if (row[j].search(/[1-8]/) !== -1) {
          var numEmptySquares = parseInt(row[j], 10)
          colIdx = colIdx + numEmptySquares
        } else {
          // piece
          var square = COLUMNS[colIdx] + currentRow
          position[square] = fenToPieceCode(row[j])
          colIdx = colIdx + 1
        }
      }

      currentRow = currentRow - 1
    }

    return position
  }

  // position object to FEN string
  // returns false if the obj is not a valid position object
  function objToFen (obj) {
    if (!validPositionObject(obj)) return false

    var fen = ''

    var currentRow = 8
    for (var i = 0; i < 8; i++) {
      for (var j = 0; j < 8; j++) {
        var square = COLUMNS[j] + currentRow

        // piece exists
        if (obj.hasOwnProperty(square)) {
          fen = fen + pieceCodeToFen(obj[square])
        } else {
          // empty space
          fen = fen + '1'
        }
      }

      if (i !== 7) {
        fen = fen + '/'
      }

      currentRow = currentRow - 1
    }

    // squeeze the empty numbers together
    fen = squeezeFenEmptySquares(fen)

    return fen
  }

  if (RUN_ASSERTS) {
    console.assert(objToFen(START_POSITION) === START_FEN)
    console.assert(objToFen({}) === '8/8/8/8/8/8/8/8')
    console.assert(objToFen({a2: 'wP', 'b2': 'bP'}) === '8/8/8/8/8/8/Pp6/8')
  }

  function squeezeFenEmptySquares (fen) {
    return fen.replace(/11111111/g, '8')
      .replace(/1111111/g, '7')
      .replace(/111111/g, '6')
      .replace(/11111/g, '5')
      .replace(/1111/g, '4')
      .replace(/111/g, '3')
      .replace(/11/g, '2')
  }

  function expandFenEmptySquares (fen) {
    return fen.replace(/8/g, '11111111')
      .replace(/7/g, '1111111')
      .replace(/6/g, '111111')
      .replace(/5/g, '11111')
      .replace(/4/g, '1111')
      .replace(/3/g, '111')
      .replace(/2/g, '11')
  }

  // returns the distance between two squares
  function squareDistance (squareA, squareB) {
    var squareAArray = squareA.split('')
    var squareAx = COLUMNS.indexOf(squareAArray[0]) + 1
    var squareAy = parseInt(squareAArray[1], 10)

    var squareBArray = squareB.split('')
    var squareBx = COLUMNS.indexOf(squareBArray[0]) + 1
    var squareBy = parseInt(squareBArray[1], 10)

    var xDelta = Math.abs(squareAx - squareBx)
    var yDelta = Math.abs(squareAy - squareBy)

    if (xDelta >= yDelta) return xDelta
    return yDelta
  }

  // returns the square of the closest instance of piece
  // returns false if no instance of piece is found in position
  function findClosestPiece (position, piece, square) {
    // create array of closest squares from square
    var closestSquares = createRadius(square)

    // search through the position in order of distance for the piece
    for (var i = 0; i < closestSquares.length; i++) {
      var s = closestSquares[i]

      if (position.hasOwnProperty(s) && position[s] === piece) {
        return s
      }
    }

    return false
  }

  // returns an array of closest squares from square
  function createRadius (square) {
    var squares = []

    // calculate distance of all squares
    for (var i = 0; i < 8; i++) {
      for (var j = 0; j < 8; j++) {
        var s = COLUMNS[i] + (j + 1)

        // skip the square we're starting from
        if (square === s) continue

        squares.push({
          square: s,
          distance: squareDistance(square, s)
        })
      }
    }

    // sort by distance
    squares.sort(function (a, b) {
      return a.distance - b.distance
    })

    // just return the square code
    var surroundingSquares = []
    for (i = 0; i < squares.length; i++) {
      surroundingSquares.push(squares[i].square)
    }

    return surroundingSquares
  }

  // given a position and a set of moves, return a new position
  // with the moves executed
  function calculatePositionFromMoves (position, moves) {
    var newPosition = deepCopy(position)

    for (var i in moves) {
      if (!moves.hasOwnProperty(i)) continue

      // skip the move if the position doesn't have a piece on the source square
      if (!newPosition.hasOwnProperty(i)) continue

      var piece = newPosition[i]
      delete newPosition[i]
      newPosition[moves[i]] = piece
    }

    return newPosition
  }

  // TODO: add some asserts here for calculatePositionFromMoves

  // ---------------------------------------------------------------------------
  // HTML
  // ---------------------------------------------------------------------------

  function buildContainerHTML (hasSparePieces, hasOverlay) {
    var html = '<div class="{chessboard}">'

    if (hasSparePieces) {
      html += '<div class="{sparePieces} {sparePiecesTop}"></div>'
    }

     if (hasOverlay === true) {
         html += '<svg class="{overlay}"></svg>';
     }

    html += '<div class="{board}"></div>'

    if (hasSparePieces) {
      html += '<div class="{sparePieces} {sparePiecesBottom}"></div>'
    }

    html += '</div>'

    return interpolateTemplate(html, CSS)
  }

  // ---------------------------------------------------------------------------
  // Config
  // ---------------------------------------------------------------------------

  function expandConfigArgumentShorthand (config) {
    if (config === 'start') {
      config = {position: deepCopy(START_POSITION)}
    } else if (validFen(config)) {
      config = {position: fenToObj(config)}
    } else if (validPositionObject(config)) {
      config = {position: deepCopy(config)}
    }

    // config must be an object
    if (!$.isPlainObject(config)) config = {}

    return config
  }

  // validate config / set default options
  function expandConfig (config) {
    // default for orientation is white
    if (config.orientation !== 'black') config.orientation = 'white'

    // default for showNotation is true
    if (config.showNotation !== false) config.showNotation = true

    // default for draggable is false
    if (config.draggable !== true) config.draggable = false

    // default for dropOffBoard is 'snapback'
    if (config.dropOffBoard !== 'trash') config.dropOffBoard = 'snapback'

    // default for sparePieces is false
    if (config.sparePieces !== true) config.sparePieces = false

    // default for sparePieces is true
    if (config.overlay !== false) config.overlay = true

    // draggable must be true if sparePieces is enabled
    if (config.sparePieces) config.draggable = true

    // default piece theme is wikipedia
    if (!config.hasOwnProperty('pieceTheme') ||
        (!isString(config.pieceTheme) && !isFunction(config.pieceTheme))) {
      config.pieceTheme = 'img/chesspieces/wikipedia/{piece}.png'
    }

    // animation speeds
    if (!validAnimationSpeed(config.appearSpeed)) config.appearSpeed = DEFAULT_APPEAR_SPEED
    if (!validAnimationSpeed(config.moveSpeed)) config.moveSpeed = DEFAULT_MOVE_SPEED
    if (!validAnimationSpeed(config.snapbackSpeed)) config.snapbackSpeed = DEFAULT_SNAPBACK_SPEED
    if (!validAnimationSpeed(config.snapSpeed)) config.snapSpeed = DEFAULT_SNAP_SPEED
    if (!validAnimationSpeed(config.trashSpeed)) config.trashSpeed = DEFAULT_TRASH_SPEED

    // throttle rate
    if (!validThrottleRate(config.dragThrottleRate)) config.dragThrottleRate = DEFAULT_DRAG_THROTTLE_RATE

    return config
  }

  // ---------------------------------------------------------------------------
  // Dependencies
  // ---------------------------------------------------------------------------

  // check for a compatible version of jQuery
  function checkJQuery () {
    if (!validJQueryVersion()) {
      var errorMsg = 'Chessboard Error 1005: Unable to find a valid version of jQuery. ' +
        'Please include jQuery ' + MINIMUM_JQUERY_VERSION + ' or higher on the page' +
        '\n\n' +
        'Exiting' + ELLIPSIS
      window.alert(errorMsg)
      return false
    }

    return true
  }

  // return either boolean false or the $container element
  function checkContainerArg (containerElOrString) {
    if (containerElOrString === '') {
      var errorMsg1 = 'Chessboard Error 1001: ' +
        'The first argument to Chessboard() cannot be an empty string.' +
        '\n\n' +
        'Exiting' + ELLIPSIS
      window.alert(errorMsg1)
      return false
    }

    // convert containerEl to query selector if it is a string
    if (isString(containerElOrString) &&
        containerElOrString.charAt(0) !== '#') {
      containerElOrString = '#' + containerElOrString
    }

    // containerEl must be something that becomes a jQuery collection of size 1
    var $container = $(containerElOrString)
    if ($container.length !== 1) {
      var errorMsg2 = 'Chessboard Error 1003: ' +
        'The first argument to Chessboard() must be the ID of a DOM node, ' +
        'an ID query selector, or a single DOM node.' +
        '\n\n' +
        'Exiting' + ELLIPSIS
      window.alert(errorMsg2)
      return false
    }

    return $container
  }

  // ---------------------------------------------------------------------------
  // Constructor
  // ---------------------------------------------------------------------------

  function constructor (containerElOrString, config) {
    // first things first: check basic dependencies
    if (!checkJQuery()) return null
    var $container = checkContainerArg(containerElOrString)
    if (!$container) return null

    // ensure the config object is what we expect
    config = expandConfigArgumentShorthand(config)
    config = expandConfig(config)

    // DOM elements
    var $board = null
    var $draggedPiece = null
    var $sparePiecesTop = null
    var $sparePiecesBottom = null
    var $overlay = null

    // constructor return object
    var widget = {}

    // -------------------------------------------------------------------------
    // Stateful
    // -------------------------------------------------------------------------

    var boardBorderSize = 2
    var currentOrientation = 'white'
    var currentPosition = {}
    var draggedPiece = null
    var draggedPieceLocation = null
    var draggedPieceSource = null
    var isDraggingPiece = false
    var draggedArrowLocation = null
    var draggedArrowSource = null
    var draggedArrowMod = null
    var annotations = [] // local storage of all annotations
    var isDraggingAnnotation = false
    // Interaction can be pieces or annotations ...
    var interactionMode = 'pieces'
    var sparePiecesElsIds = {}
    var squareElsIds = {}
    var squareElsOffsets = {}
    var squareSize = 16

    // For annotation display in svg
    var arrowSizeInStrokes = 12
    var arrowStrokesPerSquare = 6
    var circleStrokesPerSquare = 9

    // -------------------------------------------------------------------------
    // Validation / Errors
    // -------------------------------------------------------------------------

    function error (code, msg, obj) {
      // do nothing if showErrors is not set
      if (
        config.hasOwnProperty('showErrors') !== true ||
          config.showErrors === false
      ) {
        return
      }

      var errorText = 'Chessboard Error ' + code + ': ' + msg

      // print to console
      if (
        config.showErrors === 'console' &&
          typeof console === 'object' &&
          typeof console.log === 'function'
      ) {
        console.log(errorText)
        if (arguments.length >= 2) {
          console.log(obj)
        }
        return
      }

      // alert errors
      if (config.showErrors === 'alert') {
        if (obj) {
          errorText += '\n\n' + JSON.stringify(obj)
        }
        window.alert(errorText)
        return
      }

      // custom function
      if (isFunction(config.showErrors)) {
        config.showErrors(code, msg, obj)
      }
    }

    function setInitialState () {
      currentOrientation = config.orientation

      // make sure position is valid
      if (config.hasOwnProperty('position')) {
        if (config.position === 'start') {
          currentPosition = deepCopy(START_POSITION)
        } else if (validFen(config.position)) {
          currentPosition = fenToObj(config.position)
        } else if (validPositionObject(config.position)) {
          currentPosition = deepCopy(config.position)
        } else {
          error(
            7263,
            'Invalid value passed to config.position.',
            config.position
          )
        }
      }
    }

    // -------------------------------------------------------------------------
    // DOM Misc
    // -------------------------------------------------------------------------

    // calculates square size based on the width of the container
    // got a little CSS black magic here, so let me explain:
    // get the width of the container element (could be anything), reduce by 1 for
    // fudge factor, and then keep reducing until we find an exact mod 8 for
    // our square size
    function calculateSquareSize () {
      var containerWidth = parseInt($container.width(), 10)

      // defensive, prevent infinite loop
      if (!containerWidth || containerWidth <= 0) {
        return 0
      }

      // pad one pixel
      var boardWidth = containerWidth - 1

      while (boardWidth % 8 !== 0 && boardWidth > 0) {
        boardWidth = boardWidth - 1
      }

      return boardWidth / 8
    }

    // create random IDs for elements
    function createElIds () {
      // squares on the board
      for (var i = 0; i < COLUMNS.length; i++) {
        for (var j = 1; j <= 8; j++) {
          var square = COLUMNS[i] + j
          squareElsIds[square] = square + '-' + uuid()
        }
      }

      // spare pieces
      var pieces = 'KQRNBP'.split('')
      for (i = 0; i < pieces.length; i++) {
        var whitePiece = 'w' + pieces[i]
        var blackPiece = 'b' + pieces[i]
        sparePiecesElsIds[whitePiece] = whitePiece + '-' + uuid()
        sparePiecesElsIds[blackPiece] = blackPiece + '-' + uuid()
      }
    }

    // -------------------------------------------------------------------------
    // Markup Building
    // -------------------------------------------------------------------------

    function buildBoardHTML (orientation) {
      if (orientation !== 'black') {
        orientation = 'white'
      }

      var html = ''

      // algebraic notation / orientation
      var alpha = deepCopy(COLUMNS)
      var row = 8
      if (orientation === 'black') {
        alpha.reverse()
        row = 1
      }

      var squareColor = 'white'
      for (var i = 0; i < 8; i++) {
        html += '<div class="{row}">'
        for (var j = 0; j < 8; j++) {
          var square = alpha[j] + row

          html += '<div class="{square} ' + CSS[squareColor] + ' ' +
            'square-' + square + '" ' +
            'style="width:' + squareSize + 'px;height:' + squareSize + 'px;" ' +
            'id="' + squareElsIds[square] + '" ' +
            'data-square="' + square + '">'

          if (config.showNotation) {
            // alpha notation
            if ((orientation === 'white' && row === 1) ||
                (orientation === 'black' && row === 8)) {
              html += '<div class="{notation} {alpha}">' + alpha[j] + '</div>'
            }

            // numeric notation
            if (j === 0) {
              html += '<div class="{notation} {numeric}">' + row + '</div>'
            }
          }

          html += '</div>' // end .square

          squareColor = (squareColor === 'white') ? 'black' : 'white'
        }
        html += '<div class="{clearfix}"></div></div>'

        squareColor = (squareColor === 'white') ? 'black' : 'white'

        if (orientation === 'white') {
          row = row - 1
        } else {
          row = row + 1
        }
      }

      return interpolateTemplate(html, CSS)
    }

    function buildPieceImgSrc (piece) {
      if (isFunction(config.pieceTheme)) {
        return config.pieceTheme(piece)
      }

      if (isString(config.pieceTheme)) {
        return interpolateTemplate(config.pieceTheme, {piece: piece})
      }

      // NOTE: this should never happen
      error(8272, 'Unable to build image source for config.pieceTheme.')
      return ''
    }

    function buildPieceHTML (piece, hidden, id) {
      var html = '<img src="' + buildPieceImgSrc(piece) + '" '
      if (isString(id) && id !== '') {
        html += 'id="' + id + '" '
      }
      html += 'alt="" ' +
        'class="{piece}" ' +
        'data-piece="' + piece + '" ' +
        'style="width:' + squareSize + 'px;' + 'height:' + squareSize + 'px;'

      if (hidden) {
        html += 'display:none;'
      }

      html += '" />'

      return interpolateTemplate(html, CSS)
    }

    function buildSparePiecesHTML (color) {
      var pieces = ['wK', 'wQ', 'wR', 'wB', 'wN', 'wP']
      if (color === 'black') {
        pieces = ['bK', 'bQ', 'bR', 'bB', 'bN', 'bP']
      }

      var html = ''
      for (var i = 0; i < pieces.length; i++) {
        html += buildPieceHTML(pieces[i], false, sparePiecesElsIds[pieces[i]])
      }

      return html
    }

    // --------------------------------------------------------------
    //  Overlay building
    // --------------------------------------------------------------

    // Creates an svg element
    // Adapted from chessboard 0.3.0-r1
    function createSVGElement(tag, attr) {
	    // jQuery seemingly can't handle namespaces or case-sensitive attributes,
	    // so we have to go old-skool.
	    var svgEl = document.createElementNS('http://www.w3.org/2000/svg', tag);
	    for(var key in attr) {
		    svgEl.setAttribute(key, attr[key]);
	    }
	    return $(svgEl);
    }
    
    // Builds the overlay html code with markers for the arrows.
    // Adapted from chessboard 0.3.0-r1
    function buildOverlay() {
	    if(config.overlay) {
        $overlay.empty();
        var defsEl = createSVGElement('defs', {});
        // Four different markers are created (each with own color), 
        // these will be switched based on the state of the modifier keys.
        var ext = ['none', 'shift', 'alt', 'ctrl'];
        for (var i = 0; i < ext.length; ++i)
        {
		      defsEl.append(
            // Create a marker for the arrow
			      createSVGElement('marker', {
				      id: 'arrowhead'+'-'+ext[i],
				      viewBox: '0 0 10 10',
				      refX: '0',
				      refY: '5',
				      markerUnits: 'strokeWidth',
              markerWidth: arrowSizeInStrokes,
				      orient: 'auto'}).
            append(
              // This is the path for the arrowhead
              createSVGElement('path',
               {
                class: CSS['overlayArrow']+'-'+ext[i],
					      d: 'M 0 0 L 5 5 L 0 10 Z' })
			      )
		      );
          $overlay.append(defsEl);
        }
	    }
    };	

    // -------------------------------------------------------------------------
    // Animations
    // -------------------------------------------------------------------------

    function animateSquareToSquare (src, dest, piece, completeFn) {
      // get information about the source and destination squares
      var $srcSquare = $('#' + squareElsIds[src])
      var srcSquarePosition = $srcSquare.offset()
      var $destSquare = $('#' + squareElsIds[dest])
      var destSquarePosition = $destSquare.offset()

      // create the animated piece and absolutely position it
      // over the source square
      var animatedPieceId = uuid()
      $('body').append(buildPieceHTML(piece, true, animatedPieceId))
      var $animatedPiece = $('#' + animatedPieceId)
      $animatedPiece.css({
        display: '',
        position: 'absolute',
        top: srcSquarePosition.top,
        left: srcSquarePosition.left
      })

      // remove original piece from source square
      $srcSquare.find('.' + CSS.piece).remove()

      function onFinishAnimation1 () {
        // add the 'real' piece to the destination square
        $destSquare.append(buildPieceHTML(piece))

        // remove the animated piece
        $animatedPiece.remove()

        // run complete function
        if (isFunction(completeFn)) {
          completeFn()
        }
      }

      // animate the piece to the destination square
      var opts = {
        duration: config.moveSpeed,
        complete: onFinishAnimation1
      }
      $animatedPiece.animate(destSquarePosition, opts)
    }

    function animateSparePieceToSquare (piece, dest, completeFn) {
      var srcOffset = $('#' + sparePiecesElsIds[piece]).offset()
      var $destSquare = $('#' + squareElsIds[dest])
      var destOffset = $destSquare.offset()

      // create the animate piece
      var pieceId = uuid()
      $('body').append(buildPieceHTML(piece, true, pieceId))
      var $animatedPiece = $('#' + pieceId)
      $animatedPiece.css({
        display: '',
        position: 'absolute',
        left: srcOffset.left,
        top: srcOffset.top
      })

      // on complete
      function onFinishAnimation2 () {
        // add the 'real' piece to the destination square
        $destSquare.find('.' + CSS.piece).remove()
        $destSquare.append(buildPieceHTML(piece))

        // remove the animated piece
        $animatedPiece.remove()

        // run complete function
        if (isFunction(completeFn)) {
          completeFn()
        }
      }

      // animate the piece to the destination square
      var opts = {
        duration: config.moveSpeed,
        complete: onFinishAnimation2
      }
      $animatedPiece.animate(destOffset, opts)
    }

    // execute an array of animations
    function doAnimations (animations, oldPos, newPos) {
      if (animations.length === 0) return

      var numFinished = 0
      function onFinishAnimation3 () {
        // exit if all the animations aren't finished
        numFinished = numFinished + 1
        if (numFinished !== animations.length) return

        drawPositionInstant()

        // run their onMoveEnd function
        if (isFunction(config.onMoveEnd)) {
          config.onMoveEnd(deepCopy(oldPos), deepCopy(newPos))
        }
      }

      for (var i = 0; i < animations.length; i++) {
        var animation = animations[i]

        // clear a piece
        if (animation.type === 'clear') {
          $('#' + squareElsIds[animation.square] + ' .' + CSS.piece)
            .fadeOut(config.trashSpeed, onFinishAnimation3)

        // add a piece with no spare pieces - fade the piece onto the square
        } else if (animation.type === 'add' && !config.sparePieces) {
          $('#' + squareElsIds[animation.square])
            .append(buildPieceHTML(animation.piece, true))
            .find('.' + CSS.piece)
            .fadeIn(config.appearSpeed, onFinishAnimation3)

        // add a piece with spare pieces - animate from the spares
        } else if (animation.type === 'add' && config.sparePieces) {
          animateSparePieceToSquare(animation.piece, animation.square, onFinishAnimation3)

        // move a piece from squareA to squareB
        } else if (animation.type === 'move') {
          animateSquareToSquare(animation.source, animation.destination, animation.piece, onFinishAnimation3)
        }
      }
    }

    // calculate an array of animations that need to happen in order to get
    // from pos1 to pos2
    function calculateAnimations (pos1, pos2) {
      // make copies of both
      pos1 = deepCopy(pos1)
      pos2 = deepCopy(pos2)

      var animations = []
      var squaresMovedTo = {}

      // remove pieces that are the same in both positions
      for (var i in pos2) {
        if (!pos2.hasOwnProperty(i)) continue

        if (pos1.hasOwnProperty(i) && pos1[i] === pos2[i]) {
          delete pos1[i]
          delete pos2[i]
        }
      }

      // find all the "move" animations
      for (i in pos2) {
        if (!pos2.hasOwnProperty(i)) continue

        var closestPiece = findClosestPiece(pos1, pos2[i], i)
        if (closestPiece) {
          animations.push({
            type: 'move',
            source: closestPiece,
            destination: i,
            piece: pos2[i]
          })

          delete pos1[closestPiece]
          delete pos2[i]
          squaresMovedTo[i] = true
        }
      }

      // "add" animations
      for (i in pos2) {
        if (!pos2.hasOwnProperty(i)) continue

        animations.push({
          type: 'add',
          square: i,
          piece: pos2[i]
        })

        delete pos2[i]
      }

      // "clear" animations
      for (i in pos1) {
        if (!pos1.hasOwnProperty(i)) continue

        // do not clear a piece if it is on a square that is the result
        // of a "move", ie: a piece capture
        if (squaresMovedTo.hasOwnProperty(i)) continue

        animations.push({
          type: 'clear',
          square: i,
          piece: pos1[i]
        })

        delete pos1[i]
      }

      return animations
    }

    // -------------------------------------------------------------------------
    // Control Flow
    // -------------------------------------------------------------------------

    function drawPositionInstant () {
      // clear the board
      $board.find('.' + CSS.piece).remove()

      // add the pieces
      for (var i in currentPosition) {
        if (!currentPosition.hasOwnProperty(i)) continue

        $('#' + squareElsIds[i]).append(buildPieceHTML(currentPosition[i]))
      }
    }


    function drawBoard () {
      $board.html(buildBoardHTML(currentOrientation, squareSize, config.showNotation))
      drawPositionInstant()

      if (config.sparePieces) {
        if (currentOrientation === 'white') {
          $sparePiecesTop.html(buildSparePiecesHTML('black'))
          $sparePiecesBottom.html(buildSparePiecesHTML('white'))
        } else {
          $sparePiecesTop.html(buildSparePiecesHTML('white'))
          $sparePiecesBottom.html(buildSparePiecesHTML('black'))
        }
      }
    }

    function setCurrentPosition (position) {
      var oldPos = deepCopy(currentPosition)
      var newPos = deepCopy(position)
      var oldFen = objToFen(oldPos)
      var newFen = objToFen(newPos)

      // do nothing if no change in position
      if (oldFen === newFen) return

      // run their onChange function
      if (isFunction(config.onChange)) {
        var changedPosition = config.onChange(oldPos, newPos)
        if ( (changedPosition !== undefined) && 
             (typeof changedPosition === typeof position) )
            position = changedPosition
      }

      // update state
      currentPosition = position
    }

    function isXYOnSquare (x, y) {
      for (var i in squareElsOffsets) {
        if (!squareElsOffsets.hasOwnProperty(i)) continue

        var s = squareElsOffsets[i]
        if (x >= s.left &&
            x < s.left + squareSize &&
            y >= s.top &&
            y < s.top + squareSize) {
          return i
        }
      }

      return 'offboard'
    }

    // records the XY coords of every square into memory
    function captureSquareOffsets () {
      squareElsOffsets = {}

      for (var i in squareElsIds) {
        if (!squareElsIds.hasOwnProperty(i)) continue

        squareElsOffsets[i] = $('#' + squareElsIds[i]).offset()
      }
    }

    function removeCircleHighlights () {
      $board
        .find('.' + CSS.square)
        .removeClass(CSS.highlight1 + ' ' + CSS.highlight2)
    }

    function snapbackDraggedPiece () {
      // there is no "snapback" for spare pieces
      if (draggedPieceSource === 'spare') {
        trashDraggedPiece()
        return
      }

      removeCircleHighlights()

      // animation complete
      function complete () {
        drawPositionInstant()
        $draggedPiece.css('display', 'none')

        // run their onSnapbackEnd function
        if (isFunction(config.onSnapbackEnd)) {
          config.onSnapbackEnd(
            draggedPiece,
            draggedPieceSource,
            deepCopy(currentPosition),
            currentOrientation
          )
        }
      }

      // get source square position
      var sourceSquarePosition = $('#' + squareElsIds[draggedPieceSource]).offset()

      // animate the piece to the target square
      var opts = {
        duration: config.snapbackSpeed,
        complete: complete
      }
      $draggedPiece.animate(sourceSquarePosition, opts)

      // set state
      isDraggingPiece = false
    }

    function trashDraggedPiece () {
      removeCircleHighlights()

      // remove the source piece
      var newPosition = deepCopy(currentPosition)
      delete newPosition[draggedPieceSource]
      setCurrentPosition(newPosition)

      // redraw the position
      drawPositionInstant()

      // hide the dragged piece
      $draggedPiece.fadeOut(config.trashSpeed)

      // set state
      isDraggingPiece = false
    }

    function dropDraggedPieceOnSquare (square) {
      removeCircleHighlights()

      // update position
      var newPosition = deepCopy(currentPosition)
      delete newPosition[draggedPieceSource]
      newPosition[square] = draggedPiece
      setCurrentPosition(newPosition)

      // get target square information
      var targetSquarePosition = $('#' + squareElsIds[square]).offset()

      // animation complete
      function onAnimationComplete () {
        drawPositionInstant()
        $draggedPiece.css('display', 'none')

        // execute their onSnapEnd function
        if (isFunction(config.onSnapEnd)) {
          config.onSnapEnd(draggedPieceSource, square, draggedPiece)
        }
      }

      // snap the piece to the target square
      var opts = {
        duration: config.snapSpeed,
        complete: onAnimationComplete
      }
      $draggedPiece.animate(targetSquarePosition, opts)

      // set state
      isDraggingPiece = false
    }

    function beginDraggingPiece (source, piece, x, y) {
      // run their custom onDragStart function
      // their custom onDragStart function can cancel drag start
      if (isFunction(config.onDragStart) &&
          config.onDragStart(source, piece, deepCopy(currentPosition), currentOrientation) === false) {
        return
      }

      // set state
      isDraggingPiece = true
      draggedPiece = piece
      draggedPieceSource = source

      // if the piece came from spare pieces, location is offboard
      if (source === 'spare') {
        draggedPieceLocation = 'offboard'
      } else {
        draggedPieceLocation = source
      }

      // capture the x, y coords of all squares in memory
      captureSquareOffsets()

      // create the dragged piece
      $draggedPiece.attr('src', buildPieceImgSrc(piece)).css({
        display: '',
        position: 'absolute',
        left: x - squareSize / 2,
        top: y - squareSize / 2
      })

      if (source !== 'spare') {
        // highlight the source square and hide the piece
        $('#' + squareElsIds[source])
          .addClass(CSS.highlight1)
          .find('.' + CSS.piece)
          .css('display', 'none')
      }
    }

    function updateDraggedPiece (x, y) {
      // put the dragged piece over the mouse cursor
      $draggedPiece.css({
        left: x - squareSize / 2,
        top: y - squareSize / 2
      })

      // get location
      var location = isXYOnSquare(x, y)

      // do nothing if the location has not changed
      if (location === draggedPieceLocation) return

      // remove highlight from previous square
      if (validSquare(draggedPieceLocation)) {
        $('#' + squareElsIds[draggedPieceLocation]).removeClass(CSS.highlight2)
      }

      // add highlight to new square
      if (validSquare(location)) {
        $('#' + squareElsIds[location]).addClass(CSS.highlight2)
      }

      // run onDragMove
      if (isFunction(config.onDragMove)) {
        config.onDragMove(
          location,
          draggedPieceLocation,
          draggedPieceSource,
          draggedPiece,
          deepCopy(currentPosition),
          currentOrientation
        )
      }

      // update state
      draggedPieceLocation = location
    }

    function stopDraggedPiece (location) {
      // determine what the action should be
      var action = 'drop'
      if (location === 'offboard' && config.dropOffBoard === 'snapback') {
        action = 'snapback'
      }
      if (location === 'offboard' && config.dropOffBoard === 'trash') {
        action = 'trash'
      }

      // run their onDrop function, which can potentially change the drop action
      if (isFunction(config.onDrop)) {
        var newPosition = deepCopy(currentPosition)

        // source piece is a spare piece and position is off the board
        // if (draggedPieceSource === 'spare' && location === 'offboard') {...}
        // position has not changed; do nothing

        // source piece is a spare piece and position is on the board
        if (draggedPieceSource === 'spare' && validSquare(location)) {
          // add the piece to the board
          newPosition[location] = draggedPiece
        }

        // source piece was on the board and position is off the board
        if (validSquare(draggedPieceSource) && location === 'offboard') {
          // remove the piece from the board
          delete newPosition[draggedPieceSource]
        }

        // source piece was on the board and position is on the board
        if (validSquare(draggedPieceSource) && validSquare(location)) {
          // move the piece
          delete newPosition[draggedPieceSource]
          newPosition[location] = draggedPiece
        }

        var oldPosition = deepCopy(currentPosition)

        var result = config.onDrop(
          draggedPieceSource,
          location,
          draggedPiece,
          newPosition,
          oldPosition,
          currentOrientation
        )
        if (result === 'snapback' || result === 'trash') {
          action = result
        }
      }

      // do it!
      if (action === 'snapback') {
        snapbackDraggedPiece()
      } else if (action === 'trash') {
        trashDraggedPiece()
      } else if (action === 'drop') {
        dropDraggedPieceOnSquare(location)
      }
    }


    // ---------------------------------------------------
    // Functions for 
    // ---------------------------------------------------

    // Dragging is stopped, nothting is added (e.g. when
    // dragging went outside the board)
    function trashDraggedArrow () {
      removeCircleHighlights()
      // set state
      isDraggingAnnotation = false
    }

    // Called when dragging is stopped. An arrow
    // will be added, unless there is already an
    // arrow from the source to the target; in that case
    // the previous arrow will be deleted.
    function addDraggedArrow (square, mod) {
      removeCircleHighlights()

      // reset state and current
      isDraggingAnnotation = false
      removeCurrentArrow()

      // Skip drawing if the first square equals the last
      if (draggedArrowSource != square)
      // check for an existing arrow..
        if (existsArrowAnnotation(draggedArrowSource, square))
          widget.removeArrowAnnotation(draggedArrowSource, square)
        else
          widget.addArrowAnnotation(draggedArrowSource, square, mod)
    }

    function beginDraggingAnnotation (source, evt, x, y) {

      // set state
      isDraggingAnnotation = true
      draggedArrowSource = source

      // skip in case of spare or offboard
      if ( (source === 'spare') ||
           (source === 'offboard') ) {
        draggedArrowLocation = null
        draggedArrowSource = null
        isDraggingAnnotation = false
        return
      } 

      // Start dragging
      draggedArrowLocation = source
      draggedArrowMod = modFromEvent(evt)

      // capture the x, y coords of all squares in memory
      captureSquareOffsets()

      // highlight the source square and hide the piece
      $('#' + squareElsIds[source])
        .addClass(CSS.highlight1)
    }

    /*
     * is called when the mouse is dragged while 
     * drawing an arrow. Will check whether an 
     * update is required (new square), and
     * if so, update the SVG.
     */
    function updateDraggedAnnotation (x, y, mod) {

      // get location
      var location = isXYOnSquare(x, y)

      // do nothing if the location has not changed
      // and neither the modifiying keys
      // (if modifiers changed, a redraw will give the 
      // right new color)
      if ( (location === draggedArrowLocation) && 
           (mod === draggedArrowMod) )
           return;

      // remove highlight from previous square
      if (location != draggedArrowLocation) {
        if (validSquare(draggedArrowLocation)) {
          $('#' + squareElsIds[draggedArrowLocation]).removeClass(CSS.highlight2)
        }

        // add highlight to new square
        if (validSquare(location)) {
          $('#' + squareElsIds[location]).addClass(CSS.highlight2)
        }
      }

      // update current arrow
      addCurrentArrow(draggedArrowSource, location, mod)

      // update state
      draggedArrowLocation = location
      draggedArrowMod = mod
    }

    function stopDraggedAnnotation (location, mod) {
      // Reset all visual inputs
      removeCircleHighlights()
      removeCurrentArrow();

      // Remember that we stopped dragging
      isDraggingAnnotation = false

      // If location is valid square
      if (validSquare(location)) {
        // Arrow or circle, depending on whether we did move or not...
        if (location !== draggedArrowSource) 
            addDraggedArrow(location, mod)
        else 
            clickSquare(location, mod)
      }
    }

    // Checks whether an arrow exists. 
    // Checks in the local datastructure, could also be SVG (they should
    // be corresponding)
    function existsArrowAnnotation (source, target) {

      var found = []
      // Look for arrows in local datastructure
      for (var i = 0; i < annotations.length; ++i)
        if (annotations[i][0] === 'a' &&
            annotations[i][1] === source && 
            (target === undefined || annotations[i][2] === target) ) {
              found.unshift(i)
            }
      return found.length > 0;
    }

    // Current arrow is used while dragging. It has a special group
    // name (g.current), and is only added to the overlay (not to 
    // the local datastructure)
    function addCurrentArrow(source, target, mod) {

      // First remove all current arrows
      removeCurrentArrow();
      if (source === target) 
        return;

	    if(config.overlay === true) {
  	    var groupEl = createSVGElement('g', { 'class': 'current' });
        $overlay.append(groupEl);
		    
		    var pathEl = createSVGElement('path', {
          'class' :       CSS['overlayArrow'] + '-' + mod + ' current',
          'opacity': 0.5,
			    'd':            computePath(source, target),
			    'stroke-width': (squareSize / arrowStrokesPerSquare)
		    });
		    groupEl.append(pathEl);
	    }
    }
    
    // Current arrow is used while dragging. It has a special group
    // name (g.current), and needs to be removed after stopping dragging.
    // This arrow only is present in the overlay, and NOT in the local
    // datastructure.
    function removeCurrentArrow() {
	    if(config.overlay === true) {
		    $overlay.find('> g.current').remove();
	    }
    }

    // Checks whether there is a circle at the source square.
    // Check is based on svg...
    function existsCircleAnnotation (source) {

      // checks in the local datastructure 
      var found = []
      // Look for arrows in local datastructure
      for (var i = 0; i < annotations.length; ++i)
        if (annotations[i][0] === 'c' &&
            annotations[i][1] === source ) {
              found.unshift(i)
            }
      return found.length > 0;
    }

    function clickSquare(square, mod)
    {
      if ( (square === 'spare') ||
           (square === 'offboard') ) {
             return
      }

      // Default interactive behavior: remove circles when the already exist.
      if (existsCircleAnnotation(square))
        widget.removeCircleAnnotation(square)
      else
        widget.addCircleAnnotation(square, mod);
    }

    // Returns a copy of the local datastructure with
    // annotations (for persistence purposes)
    function getAnnotations() {
      return annotations;
    }


    // Restores annotations from the annos parameter.
    // (adds them, does not do a clear first!)
    // Also does not check for existing annotations that are
    // the same...
    function redrawAnnotations(annos) {

      if (annos === null || annos == undefined)
        return

      // prevent that we accidently add annotations to the
      // existing annotations.
      annos = Array.from(annos);

      // For all annotations...
      for (var i = 0; i < annos.length; ++i)
      {
        // Get the type of annotation
        var anno = annos[i][0];
        if (anno === 'a') {
          // Arrow case, get source, target, mod and redraw
          var source = annos[i][1];
          var target = annos[i][2];
          var mod = annos[i][3];
          widget.addArrowAnnotation(source, target, mod)
        }
        if (anno === 'c') {
          // Circel case, get source, mod and redraw
          var source = annos[i][1];
          var mod = annos[i][2];
          widget.addCircleAnnotation(source, mod);
        }
      }
    }

    
    // 
    function restoreAnnotations() {
      var annos = getAnnotations();
      widget.clearAnnotation();
      redrawAnnotations(annos);
    }


    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------

    // clear the board
    widget.clear = function (useAnimation) {
      widget.position({}, useAnimation)
    }

    // remove the widget from the page
    widget.destroy = function () {
      // remove markup
      $container.html('')
      $draggedPiece.remove()

      // remove event handlers
      $container.unbind()
    }

    // shorthand method to get the current FEN
    widget.fen = function () {
      return widget.position('fen')
    }

    // flip orientation
    widget.flip = function () {
      return widget.orientation('flip')
    }

    // move pieces
    // TODO: this method should be variadic as well as accept an array of moves
    widget.move = function () {
      // no need to throw an error here; just do nothing
      // TODO: this should return the current position
      if (arguments.length === 0) return

      var useAnimation = true

      // collect the moves into an object
      var moves = {}
      for (var i = 0; i < arguments.length; i++) {
        // any "false" to this function means no animations
        if (arguments[i] === false) {
          useAnimation = false
          continue
        }

        // skip invalid arguments
        if (!validMove(arguments[i])) {
          error(2826, 'Invalid move passed to the move method.', arguments[i])
          continue
        }

        var tmp = arguments[i].split('-')
        moves[tmp[0]] = tmp[1]
      }

      // calculate position from moves
      var newPos = calculatePositionFromMoves(currentPosition, moves)

      // update the board
      widget.position(newPos, useAnimation)

      // return the new position object
      return newPos
    }

    widget.interactionMode = function(arg) {
      if (arguments.length === 0) return interactionMode;

      if (arg === 'pieces' || arg === 'annotations')
          interactionMode = arg
    }

    // Sets or returns the visibility
    widget.sparePiecesVisibility = function(arg) {

      // Do nothing if no spare pieces
      if (!config.sparePieces) 
          return '';

      if (arguments.length === 0) {
          // Return visibility
          var wKelem = document.getElementById(sparePiecesElsIds['wK']);
          if (wKelem) {
            var visibility = wKelem.style.visibility
            if (visibility === '') visibility = 'visible'
            return visibility;
          } else 
            return ''
      } else {
          // Set visibility after checking value of argument
          if (arg === 'visible' || arg === 'hidden')
          {
            var visibility = arg;
            var pieces = ['wK', 'wQ', 'wR', 'wB', 'wN', 'wP',
                          'bK', 'bQ', 'bR', 'bB', 'bN', 'bP'];
            for (var i = 0; i < pieces.length; i++) {
              var id = sparePiecesElsIds[pieces[i]]
              document.getElementById(id).style.visibility = visibility
            }
          }
      }
    }


    // Toggle the visiblity of spare pieces
    widget.toggleSparePieces = function(arg) {

      // Do nothing if no spare pieces
      if (!config.sparePieces) 
          return '';


      // Get visibility
      var visibility = widget.sparePiecesVisibility();

      // Toggle visibility
      if (visibility === '' || visibility === 'visible')
          visibility = 'hidden'
      else 
          visibility = 'visible'

      // Set new visibility
      widget.sparePiecesVisibility(visibility)

    }

    widget.orientation = function (arg) {
      // no arguments, return the current orientation
      if (arguments.length === 0) {
        return currentOrientation
      }

      var sparesVisible = widget.sparePiecesVisibility()

      // set to white or black
      if (arg === 'white' || arg === 'black') {
        currentOrientation = arg
        drawBoard()
        widget.sparePiecesVisibility(sparesVisible)
        restoreAnnotations()
        return currentOrientation
      }

      // flip orientation
      if (arg === 'flip') {
        currentOrientation = currentOrientation === 'white' ? 'black' : 'white'
        drawBoard()
        widget.sparePiecesVisibility(sparesVisible)
        restoreAnnotations()
        return currentOrientation
      }

      error(5482, 'Invalid value passed to the orientation method.', arg)
    }

    widget.position = function (position, useAnimation) {
      // no arguments, return the current position
      if (arguments.length === 0) {
        return deepCopy(currentPosition)
      }

      // get position as FEN
      if (isString(position) && position.toLowerCase() === 'fen') {
        return objToFen(currentPosition)
      }

      // start position
      if (isString(position) && position.toLowerCase() === 'start') {
        position = deepCopy(START_POSITION)
      }

      // convert FEN to position object
      if (validFen(position)) {
        position = fenToObj(position)
      }

      // validate position object
      if (!validPositionObject(position)) {
        error(6482, 'Invalid value passed to the position method.', position)
        return
      }

      // default for useAnimations is true
      if (useAnimation !== false) useAnimation = true

      if (useAnimation) {
        // start the animations
        var animations = calculateAnimations(currentPosition, position)
        doAnimations(animations, currentPosition, position)

        // set the new position
        setCurrentPosition(position)
      } else {
        // instant update
        setCurrentPosition(position)
        drawPositionInstant()
      }
    }

    widget.resize = function () {
      // calulate the new square size
      squareSize = calculateSquareSize()

      // set board width
      $board.css('width', squareSize * 8 + 'px')
      $overlay.css('width', squareSize * 8 + 'px');
      $overlay.css('height', squareSize * 8 + 'px');
      $overlay.css('padding', boardBorderSize + 'px');


      // set drag piece size
      $draggedPiece.css({
        height: squareSize,
        width: squareSize
      })

      // spare pieces
      if (config.sparePieces) {
        $container
          .find('.' + CSS.sparePieces)
          .css('paddingLeft', squareSize + boardBorderSize + 'px')
      }

      // redraw the board
      drawBoard();
      restoreAnnotations();
    }

    // set the starting position
    widget.start = function (useAnimation) {
      widget.position('start', useAnimation)
    }

    // Adds an arrow annotation to the overlay, as well as
    // to the local datastructure.
    widget.addArrowAnnotation = function(source, target, mod) {

      annotations.push(['a', source, target, mod]);

      // Create a group with the start square
	    if(config.overlay === true) {
		    var groupEl = $overlay.find('> .square-' + source);
		    if(!groupEl.length) {
			    groupEl = createSVGElement('g', {
				    'class': 'square-' + source
			    });
          $overlay.append(groupEl);
        }
        
        // Create a group with the end-square, and one that defines the arrow
		    var pathEl = $overlay.find('> g.square-' + source + ' > path.square-' + target);
		    if(!pathEl.length) {
			    pathEl = createSVGElement('path', {
            'class' :       CSS['overlayArrow'] + '-' + mod + ' square-' + target,
            'opacity': 0.5,
				    'd':            computePath(source, target),
				    'stroke-width': (squareSize / arrowStrokesPerSquare)
			    });
			    groupEl.append(pathEl);
        }
	    }
    }

    // Removes an arrow annotation (both from the svg and 
    // from the local datastructure)
    // Will remove all overalys and datastructure arrows
    // that have matching source and target square.
    // No checks are being done on whether the datastructure
    // and svg overlay are matching.
    widget.removeArrowAnnotation = function(source, target) {
      // Remove from the overlay
	    if(config.overlay === true) {
		    $overlay.find(
			    '> g.square-' + source
			    + (target !== undefined ? ' > path.square-' + target : '')
		    ).remove();
      }

      // Remove arrows in local datastructure
      var found = []
      for (var i = 0; i < annotations.length; ++i)
        if (annotations[i][0] === 'a' && 
            annotations[i][1] === source && 
            annotations[i][2] === target) {
              found.unshift(i)
            }
      for (var i = 0; i < found.length; ++i)
        annotations.splice(found[i],1);
    }

    // Adds a circle annotation to the svg overlay, and 
    // to the local datastructure
    widget.addCircleAnnotation = function(source, mod) {

      // Add to local datastructure
      annotations.push(['c', source, mod]);

      // Create a group with the start square
	    if(config.overlay === true) {
		    var groupEl = $overlay.find('> .squarecircle-' + source);
		    if(!groupEl.length) {
			    groupEl = createSVGElement('g', {
				    'class': 'squarecircle-' + source
			    });
        }
        var coords = computeCoords(source)
        groupEl.append(
          // This is the circle
				  createSVGElement('circle', {
            'class': CSS['overlayCircle'] + '-' + mod,
            'stroke-width': (squareSize / circleStrokesPerSquare),
            'opacity': 0.7,
            'cx': coords[0],
            'cy': coords[1],
      		  'r': squareSize*0.85/2 })
        )
        $overlay.append(groupEl);
	    }
    }

    // Remove circle annotation from the overlay and local datastructure
    // (it is not checked whethter both are the same): all 
    // circles at the square denoted by source are removed from
    // the svg overlay, and from the local datastructure
    widget.removeCircleAnnotation = function(source ) { 

      // Overlay
	    if(config.overlay === true) {
		    $overlay.find('> g.squarecircle-' + source).remove();
      }

      // Remove circles in local datastructure: first find them...
      var found = []
      for (var i = 0; i < annotations.length; ++i)
        if (annotations[i][0] === 'c' && 
            annotations[i][1] === source) {
              found.unshift(i)
            }
      // And then remove them...
      for (var i = 0; i < found.length; ++i)
        annotations.splice(found[i],1);
    }
    
    // Clears all annotations...
    // (both local datastructure and from the svg ovelay)
    widget.clearAnnotation = function() {
      annotations = []
	    buildOverlay();
    }
    // -------------------------------------------------------------------------
    // Browser Events
    // -------------------------------------------------------------------------

    function stopDefault (evt) {
      evt.preventDefault()
    }

    /*
     * NOT USED ANYMORE
     * 
     * clickSquare is now automatically run if 
     * dragging an arrow starts at the same 
     * location as were it ended...
     * 
    function mouseauxclickSquare(evt) {
      var square = $(this).attr('data-square');
      clickSquare(square, modFromEvent(evt));
    }
    */

    function mousedownSquare (evt) {

      // Below code must be in this function ($(this)),
      // does not work in a delegated / dispatched function...?
      var square = $(this).attr('data-square')
      if (!validSquare(square)) return

      // do nothing if we're not draggable
      if (!config.draggable) return

      // Dispatch based on the button pressed, and current
      // interactionMode. LMB and 'pieces' will drag pieces,
      // otherwise annotations can be drawn.
      if (evt.button === 0 && interactionMode === 'pieces') {
        mousedownSquareForPiece (square, evt)
      } else
        mousedownSquareForAnnotation (square, evt)
    }

    function mousedownSquareForPiece(square, evt)
    {
      // do nothing if there is no piece on this square
      if (!currentPosition.hasOwnProperty(square)) return
      beginDraggingPiece(square, currentPosition[square], evt.pageX, evt.pageY)
    }

    function mousedownSquareForAnnotation (square, evt) {
      beginDraggingAnnotation(square, evt, evt.pageX, evt.pageY)
    }

    /*
     * Functions for debugging, use an html element called "logger" to 
     * output text.
     */
    function ui_log(text)
    {
      var logger = document.getElementById("logger")
      if (logger) {
          logger.value += text + "\n"
          logger.scrollTop = logger.scrollHeight
      }
    }

    var e_counter = 0;
    /*
     * Function used for debugging the touch events...
     */ 
    function printTouchEvent(info, evt)
    {
      var evt_orig = evt.originalEvent;

      ui_log( info + "[" + e_counter + "]" );
      e_counter = e_counter + 1;
      ui_log("--#ct = " + evt_orig.changedTouches.length);
      for (var i = 0; i < evt_orig.changedTouches.length; ++i)
      {
          ui_log( "----["+i+"] -> id = " + 
               evt_orig.changedTouches[i].identifier + 
               ", pos=(" + Math.floor(evt_orig.changedTouches[i].pageX) + "," + 
               Math.floor(evt_orig.changedTouches[i].pageY) );
      }
      ui_log( "--#t = " + evt_orig.touches.length);
      for (var i = 0; i < evt_orig.touches.length; ++i)
      {
          ui_log( "----["+i+"] -> id = " + 
               evt_orig.touches[i].identifier + 
               ", pos=(" + Math.floor(evt_orig.touches[i].pageX) + "," + 
               Math.floor(evt_orig.touches[i].pageY) );
      }
    }

    function touchstartSquare (evt) {
      // do nothing if we're not draggable
      if (!config.draggable) return

      var square = $(this).attr('data-square')
      if (!validSquare(square)) return
      
      // Event is on square, thus do not process mouse events ...
      evt.preventDefault()

      var evt_orig = evt.originalEvent

      // TO DO: other ways for discriminating between 
      // dragging pieces and doing annotatoins...?
      if ( interactionMode === 'pieces') {
          // do nothing if there is no piece on this square
          if (!currentPosition.hasOwnProperty(square)) 
            return
          beginDraggingPiece(
            square,
            currentPosition[square],
            evt_orig.changedTouches[0].pageX,
            evt_orig.changedTouches[0].pageY
          )
      } else {
          beginDraggingAnnotation(square, evt, 
             evt_orig.changedTouches[0].pageX, 
             evt_orig.changedTouches[0].pageY)
      }
    }

    function mousedownSparePiece (evt) {
      // do nothing if sparePieces is not enabled
      if (!config.sparePieces) return

      // only respond to first button
      if (evt.button !== 0) return

      var piece = $(this).attr('data-piece')

      beginDraggingPiece('spare', piece, evt.pageX, evt.pageY)
    }

    function touchstartSparePiece (e) {
      // do nothing if sparePieces is not enabled
      if (!config.sparePieces) return

      var piece = $(this).attr('data-piece')

      e = e.originalEvent
      beginDraggingPiece(
        'spare',
        piece,
        e.changedTouches[0].pageX,
        e.changedTouches[0].pageY
      )
    }

    function mousemoveWindow (evt) {
      // Dispatching is based on the current dragging action.
      // This dragging action is set in the  mousedown event function.
      if (isDraggingPiece)
        updateDraggedPiece(evt.pageX, evt.pageY)
      if (isDraggingAnnotation)
        updateDraggedAnnotation(evt.pageX, evt.pageY, 
                                modFromEvent(evt))
    }

    // Checks the modifiers keys, 
    // on these the color is determined
    function modFromEvent(evt)
    {
      var mod = 'none'
      if (evt.shiftKey) 
        mod = 'shift';
      else if (evt.altKey)
        mod = 'alt';
      else if (evt.ctrlKey)
        mod = 'ctrl';

      return mod;
    }

    // Color is determined on number of
    // fingers on the screen ...
    function modFromTouchMoveEvent(evt)
    {
      // Move event: the touches contain the changed touch...
      // Count the additional touches ...
      switch (evt.touches.length) {
        case 1: return 'none';
        case 2: return 'alt';
        case 3: return 'ctrl';
        case 4: return 'shift'
        default: return 'none';
      }
    }

    function modFromTouchEndEvent(evt)
    {
      // Move event: the touches do NOT contain the changed touch...
      // Count the additional touches...
      switch (evt.touches.length) {
        case 0: return 'none';
        case 1: return 'alt';
        case 2: return 'ctrl';
        case 3: return 'shift'
        default: return 'none';
      }
    }

    var throttledMousemoveWindow = throttle(mousemoveWindow, config.dragThrottleRate)

    function touchmoveWindow (evt) {
      // prevent screen from scrolling
      evt.preventDefault()

      // do nothing if we are not dragging a piece
      if (!isDraggingPiece && !isDraggingAnnotation) return

      var evt_orig = evt.originalEvent

      // Dispatching is based on the current dragging action.
      // This dragging action is set in the  mousedown event function.
      if ( isDraggingPiece ) {
          updateDraggedPiece(evt_orig.changedTouches[0].pageX,
                             evt_orig.changedTouches[0].pageY)
      } 
      if ( isDraggingAnnotation ) {
          updateDraggedAnnotation(evt_orig.changedTouches[0].pageX, 
                                  evt_orig.changedTouches[0].pageY, 
                                  modFromTouchMoveEvent(evt_orig))
      }
    }

    var throttledTouchmoveWindow = throttle(touchmoveWindow, config.dragThrottleRate)

    function mouseupWindow (evt) {
      var location = isXYOnSquare(evt.pageX, evt.pageY)

      // Dispatch based on the dragging action.
      if (isDraggingPiece)
        stopDraggedPiece(location)
      if (isDraggingAnnotation) 
        stopDraggedAnnotation(location, modFromEvent(evt))
    }

    function touchendWindow (evt) {

      // do nothing if we are not dragging a piece
      if (!isDraggingPiece && !isDraggingAnnotation) return

      // don't do mouse events, as the touch will interprete these...
      evt.preventDefault()

      // get the location
      var location = 
          isXYOnSquare(evt.originalEvent.changedTouches[0].pageX,
                       evt.originalEvent.changedTouches[0].pageY)

      // Dispatch based on the dragging action.
      if (isDraggingPiece) 
        stopDraggedPiece(location)
      if (isDraggingAnnotation)  
        stopDraggedAnnotation(location, modFromTouchEndEvent(evt.originalEvent))
    }

    function mouseenterSquare (evt) {
      // do not fire this event if we are dragging a piece
      // NOTE: this should never happen, but it's a safeguard
      if (isDraggingPiece || isDraggingAnnotation)  return

      // exit if they did not provide a onMouseoverSquare function
      if (!isFunction(config.onMouseoverSquare)) return

      // get the square
      var square = $(evt.currentTarget).attr('data-square')

      // NOTE: this should never happen; defensive
      if (!validSquare(square)) return

      // get the piece on this square
      var piece = false
      if (currentPosition.hasOwnProperty(square)) {
        piece = currentPosition[square]
      }

      // execute their function
      config.onMouseoverSquare(square, piece, deepCopy(currentPosition), currentOrientation)
    }

    function mouseleaveSquare (evt) {
      // do not fire this event if we are dragging a piece
      // NOTE: this should never happen, but it's a safeguard
      if (isDraggingPiece || isDraggingAnnotation) return

      // exit if they did not provide an onMouseoutSquare function
      if (!isFunction(config.onMouseoutSquare)) return

      // get the square
      var square = $(evt.currentTarget).attr('data-square')

      // NOTE: this should never happen; defensive
      if (!validSquare(square)) return

      // get the piece on this square
      var piece = false
      if (currentPosition.hasOwnProperty(square)) {
        piece = currentPosition[square]
      }

      // execute their function
      config.onMouseoutSquare(square, piece, deepCopy(currentPosition), currentOrientation)
    }

    // -------------------------------------------------------------------------
    // Initialization
    // -------------------------------------------------------------------------

    function addEvents () {
      // prevent "image drag"
      $('body').on('mousedown mousemove', '.' + CSS.piece, stopDefault)

      // mouse drag pieces
      $board.on('mousedown', '.' + CSS.square, mousedownSquare)
      $container.on('mousedown', '.' + CSS.sparePieces + ' .' + CSS.piece, mousedownSparePiece)

      // mouse enter / leave square
      $board
        .on('mouseenter', '.' + CSS.square, mouseenterSquare)
        .on('mouseleave', '.' + CSS.square, mouseleaveSquare)
      
      // RMB click 
      // $board
        // .on('auxclick', '.' + CSS.square, mouseauxclickSquare) 
        // .on('click', '.' + CSS.square, mouseauxclickSquare) 

      // piece drag
      var $window = $(window)
      $window
        .on('mousemove', throttledMousemoveWindow)
        .on('mouseup', mouseupWindow)

      // touch drag pieces
      if (isTouchDevice()) {
        $board.on('touchstart', '.' + CSS.square, touchstartSquare)
        $container.on('touchstart', '.' + CSS.sparePieces + ' .' + CSS.piece, touchstartSparePiece)
        $window
          .on('touchmove', throttledTouchmoveWindow)
          .on('touchend', touchendWindow)
      }
    }

    function initDOM () {
      // create unique IDs for all the elements we will create
      createElIds()

      // build board and save it in memory
      $container.html(buildContainerHTML(config.sparePieces, config.overlay))
      $board = $container.find('.' + CSS.board)

      if (config.overlay) {
        $overlay = $container.find('.' + CSS.overlay);
        buildOverlay();
      }

      if (config.sparePieces) {
        $sparePiecesTop = $container.find('.' + CSS.sparePiecesTop)
        $sparePiecesBottom = $container.find('.' + CSS.sparePiecesBottom)
      }

      // create the drag piece
      var draggedPieceId = uuid()
      $('body').append(buildPieceHTML('wP', true, draggedPieceId))
      $draggedPiece = $('#' + draggedPieceId)

      // TODO: need to remove this dragged piece element if the board is no
      // longer in the DOM

      // get the border size
      boardBorderSize = parseInt($board.css('borderLeftWidth'), 10)

      // set the size and draw the board
      widget.resize()
    }

    
    //-------------------------------------------------------------------------
    // SVG Overlay
    //-------------------------------------------------------------------------
    function computeCoords(s1) {
	    
	    var pos = { x: COLUMNS.indexOf(s1[0]), y: parseInt(s1[1], 10) -1 };
	    
	    if(currentOrientation === 'white') {
		    pos.y = 7 - pos.y;
      } else if (currentOrientation === 'black') {
		    pos.x = 7 - pos.x;
      }
      
      return [ squareSize * (pos.x + 0.5) , squareSize * (pos.y + 0.5) ];
    }
    

    /* 
     * Construct an svg path string for an arrow drawn between two points.
     */
    function computePath(s1, s2) {
	    
	    var start = { x: COLUMNS.indexOf(s1[0]), y: parseInt(s1[1], 10) -1 };
	    var end   = { x: COLUMNS.indexOf(s2[0]), y: parseInt(s2[1], 10) -1 };
	    
	    if(currentOrientation === 'white') {
		    start.y = 7 - start.y;
        end.y   = 7 - end.y;
      } else if (currentOrientation === 'black') {
		    start.x = 7 - start.x;
		    end.x   = 7 - end.x;
	    }
	    
      // Start at the center
      var path = ['M', squareSize * (start.x + 0.5), 
                       squareSize * (start.y + 0.5)];

      // Compute step to end point.
      var dx = end.x - start.x;
      var dy = end.y - start.y;
      var len = Math.sqrt(dx*dx+dy*dy);

      // This is by empirics, and should be improved.
      // Purpose is to compute the length of the arrowhead.
      // Then that we can position e.g. the center of the
      // arrowhead at the center of the square.
      var arrowLen = arrowSizeInStrokes/(arrowStrokesPerSquare*6.0);
      // Using 0.0 instead of 0.5 below (i.e. making fraction equal to 1)
      // will give an arrowhead that starts (has its base) at the 
      // center of the image.
      var fraction;
      if (len > 0.0)
        fraction = (len - 0.5 * arrowLen) / len;
      else
        fraction = 1.0;
      
      // Update dx and dy based on the updated length.
      // (i.e. make the length a bit shorter, based on 
      // fraction)
      dx = fraction*dx;
      dy = fraction*dy;

      // Add the endpoint by using the length
      path.push('L', squareSize * (start.x + 0.5 + dx), 
                     squareSize * (start.y + 0.5 + dy));
	    
	    return path.join(' ');
    }
    
    

    // -------------------------------------------------------------------------
    // Initialization
    // -------------------------------------------------------------------------

    setInitialState()
    initDOM()
    addEvents()

    // return the widget object
    return widget
  } // end constructor

  // TODO: do module exports here
  window['Chessboard'] = constructor

  // support legacy ChessBoard name
  window['ChessBoard'] = window['Chessboard']

  // expose util functions
  window['Chessboard']['fenToObj'] = fenToObj
  window['Chessboard']['objToFen'] = objToFen
})() // end anonymous wrapper
