// This file contains the Javascript implementation of a schaakbord. 
// It defines the scripts belonging to the UserInterface that is define in schaakbord.html,
// and it used the chessboard as defined in chessboard...js
//
// The main purpose of the functions in this file are to:
// * link actions to buttons
// * allow for addition (and editing) of positions in a list of positions
// * allow for addition (and editing) of moves for a position
// * allow saving the position and moves to a file, and reloading them from
//   file.
//

// Chessboard configuration
var config = {
    draggable: true,
    dropOffBoard: 'trash',
    sparePieces: true,
    onChange: onChange
}

// Create chessboard
var board = ChessBoard('board', config);

// Create a dummy chessboard. This board is being used to 
// 'play' moves and in that way get the resulting position
// after the moves.
// This is a.o. used for checkAndDoMove()
var dummyboard = ChessBoard('dummy', { } );
// The dummy board should NOT be on the screen
document.getElementById('dummy').style.display = 'none';

// https://www.w3schools.com/howto/howto_js_fullscreen.asp
function openFullscreen() {
  var elem = document.documentElement;
  if (elem.requestFullscreen) {
    elem.requestFullscreen();
  } else if (elem.mozRequestFullScreen) { /* Firefox */
    elem.mozRequestFullScreen();
  } else if (elem.webkitRequestFullscreen) { /* Chrome, Safari and Opera */
    elem.webkitRequestFullscreen();
  } else if (elem.msRequestFullscreen) { /* IE/Edge */
    elem.msRequestFullscreen();
  }
}

/* Close fullscreen */
function closeFullscreen() {
  if (document.exitFullscreen) {
    document.exitFullscreen();
  } else if (document.mozCancelFullScreen) { /* Firefox */
    document.mozCancelFullScreen();
  } else if (document.webkitExitFullscreen) { /* Chrome, Safari and Opera */
    document.webkitExitFullscreen();
  } else if (document.msExitFullscreen) { /* IE/Edge */
    document.msExitFullscreen();
  }
}

// Toggles fullscreenmode.
function toggleFullscreen() 
{
    if (!window.screenTop && !window.screenY) {
        // Full screen
        closeFullscreen();
    } else {
        openFullscreen();
    }
}

// Toggles visibility of the left and right column (but not
// the bottom right buttons).
function togglePanels() 
{
    // Check current state (left column is taken as reference)
    if ( document.getElementById('left_col').style.visibility === 'hidden' ) {
        // adapt visibility...
        document.getElementById('left_col').style.visibility = 'visible'
        document.getElementById('right_col').style.visibility = 'visible'
    } else {
        document.getElementById('left_col').style.visibility = 'hidden'
        document.getElementById('right_col').style.visibility = 'hidden'
    }
}


// Finds out a (basic) representation of the move, based on the old position
// and the new position. 
// Basic representation means: fromsquare-tosquare.
function findMoveBetweenPositions(oldPos, newPos)
{
    // find out the start square of the move ...
    var start = []
    for (var square in oldPos) {
        // look for squares in the old position where a piece has been (re)moved from
        if (newPos[square] === undefined || newPos[square] === '')
            start.push(square)
    }

    // find out the end square of the move ...
    var end = []
    for (var square in newPos) {
        // look for square in the new position where pieces have 
        if (oldPos[square] === undefined || oldPos[square] === '' || oldPos[square] != newPos[square])
            end.push(square)
    }

    // Interpret the results...
    if (start.length === 0 && end.length === 1) {
      // No square became empty, but one square has a changed piece 
      // (or a piece has been moved there
      // Could thus be minor promotion ..., 
      // as only one square (and the same for both) has changed
      if ( (oldPos[end[0]] != undefined) && 
            oldPos[end[0]].slice(-1).match(/[QRBN]/) && 
           (end[0][1] === '1' || end[0][1] === '8') )
      {
          // Likely previous move was a minor promotion, check the current move list
          // for a promotion as last move
          var moves = document.getElementById('moves').value
          if (moves.length > 0 && moves.slice(-1).match(/[DTLP]/) )
          {
             moves = moves.slice(0,-1)+pieceBoardToUI[newPos[end[0]][1]]
             document.getElementById('moves').value = moves;
          }
      }
    }
    if (start.length == 1 && end.length == 1) {
      // normal case: one start and one end square.
      return start + '-' + end
    }
    // Other cases we do not interpret. Note that start=2 and end=1 
    // could be en passant, and start=2 and end=2 could be castling
    return ''
}


// Is called when the position on the board changes.
// When the board changes, it will find out the move that has been performed,
// if register_moves is set.
// If register_moves is set, a nice text representation is added to the 
// 'moves' parameter, and if it is a special move 
// (castling, en passent, promotion), the move itself (i.e. the board position
// will be finalized automatically (promotion is always Queen promotion).
function onChange (oldPos, newPos) {

  // Check whether the current position has changed...

  var returnpos = null;


  // Find out what has been changed ..
  var move = findMoveBetweenPositions(oldPos, newPos);
  if (move != '') {

      // Use the dummyboard to test the move....
      dummyboard.position(oldPos);
      // Do the move on the dummy board
      var nicemove = checkAndDoMove(dummyboard, move);
      // It could happen the the new position in the dummyboard is 
      // different from the newPos given. This is in the following cases:
      // * castling (when king castles, the rook is moved as well)
      // * en passent moves: the fiendly pawn is additionally removed
      // * promotion: a Queen is added to the promotion square. Minor promotion 
      //   is not supported yet. 

      // Add the move to the text representation
      if (document.getElementById('register_moves').checked ) {

          nicemove = nicemove.trim();
          if (document.getElementById('moves').value === '')
              document.getElementById('moves').value = nicemove
          else
              document.getElementById('moves').value += ' ' + nicemove
      }

      // If the position should be updated...
      if (dummyboard.position() != newPos)
      
        // Return it
        returnpos =  dummyboard.position()
  } 

  // Now check whether anything has changed w.r.t. the 'official'
  // position
  if (returnpos) 
      dummyboard.position(returnpos)
   else
      dummyboard.position(newPos)

  // Get fen of the board (after the change), and 
  var newfen = dummyboard.fen();
  // get fen of the board as it should be if no changes would be applied.
  var curfen = getFENCurrentPosition(0);

  // If new position does not match the expected one
  if (newfen !== curfen)
  {
    // Check whether the next move has been manually done ...
    var nextfen = getFENCurrentPosition(1);
    // give a color to the status line, indicating that (manual) changes have happend.
    if (newfen !== nextfen)
        // no, we did not do so..
        document.getElementById('status_position').style.color = "#008CBA"
    else {
        // Yes, the change corresponds to the manual move, update move and status.
        var dropdown = document.getElementById('movelist');
        dropdown.selectedIndex = dropdown.selectedIndex+1
        updateStatus()
        document.getElementById('status_position').style.color = "black"
    }
  } else {
    // reset the color of the status line
    document.getElementById('status_position').style.color = "black"
  }

  if (returnpos)
    return returnpos;
}

// Returns the FEN as it should be for the current position
function getFENCurrentPosition(moveoffset)
{
    // 3 Cases:
    // * no positions present or selected --> return FEN of  the board
    // * position selected, no move present or selected --> return FEN 
    //   of the position
    // * position and move selected --> return FEN of the position and
    //   next move. If the current move is the last move, use the dummy
    //   board to compute the FEN ...

    var posdropdown = document.getElementById('position_list');
    if (posdropdown.length === 0 || posdropdown.selectedIndex < 0)
        return board.fen();

    var selectedpos = posdropdown.selectedIndex;

    // there is at least a position selected
    var movedropdown = document.getElementById('movelist');
    if (movedropdown.length === 0 || movedropdown.selectedIndex + moveoffset < 0) {
        // No moves selected --> return fen of position
        var fenstring = fenfromlistvalue(posdropdown.options[selectedpos].value)
        return fenstring;
    }

    var selectedmove = movedropdown.selectedIndex + moveoffset;

    // there is also a move selected
    if (selectedmove < movedropdown.length-1) {
        // Not the last move, so return the fen from the next move
        var fen = movedropdown.options[selectedmove+1].value
        return fen.trim()
    } else if (selectedmove === movedropdown.length-1) {
        // Last move, so perform the move on the dummyboard...
        var fen = movedropdown.options[selectedmove].value
        var move = getStandardMove(movedropdown.options[selectedmove].text)
        dummyboard.position(fen);
        checkAndDoMove(dummyboard, move);
        // dummyboard.move(move);
        return dummyboard.fen();
    }
    return ''

}


// Resize the board if the window is resized.
$(window).bind('resize', function(e)
{
    // Visibility is not retained after resize (?!)
    var v = board.sparePiecesVisibility()
    board.resize()
    board.sparePiecesVisibility(v)
});

// Updates the board based on the current FEN.
function fenToBoard()
{
    var fenstring = document.getElementById('inputfen').value
    board.position(fenstring, false); 
}

// Update the current FEN based on the current position
// on the board.
function boardToFEN()
{
    var fenstring = board.fen();
    document.getElementById('inputfen').value = fenstring
}

// Prepare the elements (visibility, initialization) that
// allow pattern-matching based search for positions (on names)
function statusPositionClicked()
{
    // Reset the elements ...
    var position = ''
    var pos_list = document.getElementById('position_list')
    if (pos_list.length > 0 && pos_list.selectedIndex > -1)
        position = pos_list.options[pos_list.selectedIndex].text

    document.getElementById('search_status_position').value = position
    clearList('match_list');
    // Active the right elements ...
    document.getElementById('status_position').style.display = 'none';
    document.getElementById('search_status_position').style.display = 'inline-block';
    document.getElementById('match_list').style.display = 'inline-block';
    document.getElementById('search_status_position').focus();
    // Fills the dropdown list with the current matching positions..
    var namesdropdown = document.getElementById('position_list');
    var regexp = RegExp(position)
    for (var i = 0; i < namesdropdown.length; ++i) {
        if (namesdropdown.options[i].text.match(regexp))
            addOptionToList('match_list', namesdropdown.options[i].text, '')
        }
}

// Reset the visibility of the elements that allow pattern-matched
// search for positions.
function statusPositionReset()
{
    document.getElementById('status_position').style.display = 'inline-block'
    document.getElementById('search_status_position').style.display = 'none'
    document.getElementById('match_list').style.display = 'none'
}

// When an option in the match-list is selected, this option is used.
function matchListChanged()
{
    var dropdown = document.getElementById('match_list');
    if (dropdown.length > 0 && dropdown.selectedIndex > -1)
    {
	var name = dropdown.options[dropdown.selectedIndex].text;
        searchPositionWithName(name);
	newPositionSelected();
    }
    statusPositionReset();
}


/*
 * Functions for the User Interface
 */

 // The buttons for going through the moves and positions
 // are only visible if there are positions. 
 // This function should be called everywhere the position
 // list may be changed / updated.
function updateMoveAndPositionButtonsVisibility()
{
    // Only show buttons if there are positions ...
    if (document.getElementById('position_list').options.length > 0) {
      // turn on buttons for position and moves
      document.getElementById('position_buttons').style.display = 'inline-flex';
      document.getElementById('move_buttons').style.display = 'inline-flex';
    } else {
      // turn off buttons for position and moves
      document.getElementById('position_buttons').style.display = 'none';
      document.getElementById('move_buttons').style.display = 'none';
    }

}

// Called when the register_moves status is changed.
// Updates the visual representation of the record button.
function registerMovesStatusChanged()
{
    var checked = document.getElementById('register_moves').checked
    if (checked) {
        document.getElementById('record').style.color = 'red'
    } else {
        document.getElementById('record').style.color = 'black'
    }
}

// Update (flip) the interaction mode, from pieces to/from annotations.
function interactionModeClicked()
{
    if (board.interactionMode() === 'annotations') {
        board.interactionMode('pieces')
        document.getElementById('interactionMode').innerHTML = '[&nbsp;]&nbsp;'
    } else {
        board.interactionMode('annotations')
        document.getElementById('interactionMode').innerHTML = '[A]&nbsp;'
    }
}

// Record button is clicked: change the register_moves status
function recordClicked()
{
    var checked = document.getElementById('register_moves').checked;
    checked = !checked;
    document.getElementById('register_moves').checked = checked;
    registerMovesStatusChanged();
}

// Load a file.
function statusFileNameClicked()
{
    document.getElementById('loadfile').click()
}

/*
 * Functions for changing the position and move lists
 */

// Add current position to the position list,
// based on the current position on the board.
// A name is added, and no moves are added
// If the name already exists (or no name is given),
// that position is selected, but nothing is added.
function addPosition()
{
    var label = document.getElementById('position_name').value
    if (!searchPositionWithName() && label !== '') {
        // If a new valid name...
        // add the position
        var fenstring = board.fen();
        var dropdown = document.getElementById('position_list');
        var option = document.createElement('option')
        option.text = label
        option.value = fenstring
        dropdown.add(option)
        dropdown.selectedIndex = dropdown.length-1
        updateMoveAndPositionButtonsVisibility()
        newPositionSelected()
    }
}

// Removes the position with name (with moves) from the list
function removePosition()
{
    if (searchPositionWithName()) {

        // Do not record position changes...
        document.getElementById('register_moves').checked  = false
        registerMovesStatusChanged()

        // And the item is being removed.
        var dropdown = document.getElementById('position_list');
        var index = dropdown.selectedIndex
        dropdown.remove(dropdown.selectedIndex)
        // Update selected index...
        if (index < dropdown.options.length)
          dropdown.selectedIndex = index;
        else
          dropdown.selectedIndex = index-1;
        updateMoveAndPositionButtonsVisibility()
        // Update the board to the new position.
        newPositionSelected();
    }
}


// Update (change) current position to the position list, 
// based on the current position on the board, and the
// name specified, and the current moves.
// (i.e. moves are not touched!)
function updatePosition()
{
    var dropdown = document.getElementById('position_list');

    // No options, or none selected...
    if (dropdown.options.length === 0 ||
        dropdown.selectedIndex < 0) {
        addPosition()
        return;
    }

    // Updates can also change the name, thus we prevent duplicate names
    var newname = document.getElementById('position_name').value
    // Do not process empty names...
    if (newname === '')
        return;
    // If the name did not change, or the new name does not exist..
    if (dropdown.options[dropdown.selectedIndex].text == newname || 
        !searchPositionWithName()) {
        // perform the update...
        document.getElementById('register_moves').checked  = false
        registerMovesStatusChanged()
        // var fenstring = document.getElementById('inputfen').value
        var fenstring = board.fen();
        // Do not update the moves, as they have their own way 
        // to be updated. This only changes the (initial) position
        // var moves = document.getElementById('moves').value
        selected = dropdown.selectedIndex
        var prev_moves = movefromlistvalue(dropdown.options[selected].value);
        var option = document.createElement('option')
        option.text = newname
        option.value = fenstring  + ' # ' + prev_moves
        dropdown.options[selected] = option
        dropdown.selectedIndex = selected
        newPositionSelected()
    }
}

// After having added moves, the moves are 'filtered', and appear
// clean in the move list. If there were typos/errors in the moves
// textbox, these will still be in the movestring stored with the 
// position. Therefore, after having interpreted the moves 
// we will copy them back to the position
function updatePositionMovesString() {

    // Get a string representation of the moves 
    var moves = ''
    var dropdown = document.getElementById('movelist')
    for (var i = 0; i < dropdown.length; ++i)
        moves += stripMoveDecoration(dropdown.options[i].text) + ' '
    moves = moves.trim()

    // And store them in the position_list
    dropdown = document.getElementById('position_list')
    var selected = dropdown.selectedIndex
    var fenstring = fenfromlistvalue(dropdown.options[selected].value)
    var option = document.createElement('option')
    option.text = dropdown.options[selected].text;
    option.value = fenstring + ' # ' + moves
    dropdown.options[selected] = option
    dropdown.selectedIndex = selected
}

// Adds the current series of moves to the current position.
// Does NOT update the label of the position, or the position itself
function addMoves()
{
    var dropdown = document.getElementById('position_list');
    selected = dropdown.selectedIndex
    // If nothing selected, we do not know wheree to add...
    if (selected < 0) 
      return
      // Get the info from the UI and the position list
    var label = dropdown.options[selected].text
    var fenstring = fenfromlistvalue(dropdown.options[selected].value);
    var prev_moves = movefromlistvalue(dropdown.options[selected].value); 
    var moves = document.getElementById('moves').value
    // Create a new option, with the current info and the new moves
    var option = document.createElement('option')
    option.text = label
    option.value = fenstring + ' # ' + prev_moves + ' ' + moves
    // Replace current option
    dropdown.options[selected] = option
    dropdown.selectedIndex = selected

    // This will make sure that the new moves are put into the move list
    // Moves will be checked and filtered in this process
    newPositionSelected()

    // And this will update the move representation in the UI, 
    updatePositionMovesString() 

    // And select the last move, to be prepared for subsequent moves.
    var dropdown = document.getElementById('movelist');
    dropdown.selectedIndex = dropdown.length-1
    newMoveSelected()
}

// Removes the current series of moves to the current position.
// Does NOT update the label of the position, or the position itself
function removeMoves()
{
    var dropdown = document.getElementById('position_list');
    selected = dropdown.selectedIndex
    if (selected < 0) 
      return

    // Reset moves square
    var label = dropdown.options[selected].text
    var fenstring = fenfromlistvalue(dropdown.options[selected].value);
    var option = document.createElement('option')
    option.text = label
    // Forget the moves ...
    option.value = fenstring + ' # ' // + moves
    // Update option
    dropdown.options[selected] = option
    // Keep selection
    dropdown.selectedIndex = selected
    // Clear the movelist dropdown
    clearList('movelist')

    // This will update the rest.
    newMoveSelected()
}

// Updates the current series of moves to the current position.
// Does NOT update the label of the position, or the position itself
function updateMoves()
{
    var dropdown = document.getElementById('position_list');
    selected = dropdown.selectedIndex
    if (selected < 0) 
      return

    // Get current values
    var label = dropdown.options[selected].text
    var fenstring = fenfromlistvalue(dropdown.options[selected].value);
    var moves = document.getElementById('moves').value
    // Create new option
    var option = document.createElement('option')
    option.text = label
    option.value = fenstring + ' # ' + moves
    // Replace current option with new one
    dropdown.options[selected] = option
    // Maintain selection
    dropdown.selectedIndex = selected

    // This will update the UI appropriately
    // and also build the moves dropdown.
    newPositionSelected()
    updatePositionMovesString() 

    // Select the last move
    var dropdown = document.getElementById('movelist');
    dropdown.selectedIndex = dropdown.length-1
    // And update UI accordingly.
    newMoveSelected()
}


/*
 * Select positions / moves
 */

// Move the board (and the position list) to the first position
// If a new position could be selected, the annotations are 
// also cleared.
function firstPosition()
{
    var dropdown = document.getElementById('position_list');
    if (dropdown.length > 0) {
        dropdown.selectedIndex = 0;
        newPositionSelected()
    }
}

// Move the board (and the position list) to the previous position, if 
// possible. 
// If a new position could be selected, the annotations are 
// also cleared.
function previousPosition()
{
    var dropdown = document.getElementById('position_list');
    if (dropdown.selectedIndex > 0) {
        dropdown.selectedIndex = dropdown.selectedIndex-1
        newPositionSelected()
    }
}

// Move the board (and the position list) to the next position, if 
// possible. 
// If a new position could be selected, the annotations are 
// also cleared.
function nextPosition()
{
    var dropdown = document.getElementById('position_list');
    if (dropdown.selectedIndex < dropdown.length-1) {
        dropdown.selectedIndex = dropdown.selectedIndex+1
        // Will reset all values for the position, and clear
        // annotations.
        newPositionSelected()
    }
}

// Move the board (and the position list) to the last position, if 
// possible. 
// If a new position could be selected, the annotations are 
// also cleared.
function lastPosition()
{
    var dropdown = document.getElementById('position_list');
    if (dropdown.length > 0) {
        dropdown.selectedIndex = dropdown.length-1;
        newPositionSelected()
    }
}

// Move the board to the current position *before* the first move
// Note that this does not only perform the move, but 
// also restores any other changes on the board.
// It does NOT change the annotations.
function beforeFirstMove()
{
    // Is the same as going the (initial) current position
    resetCurrentPosition();
}

// Move the board to the position *after* the previous move
// Note that this does not only perform the move, but 
// also restores any other changes on the board.
// It does NOT change the annotations.
function previousMove()
{
    var dropdown = document.getElementById('movelist');
    if (dropdown.selectedIndex >= 0) {
        // Got to the position before the current move, if possible
        dropdown.selectedIndex = dropdown.selectedIndex-1
        board.position(dropdown.options[dropdown.selectedIndex+1].value)
    }
    updateStatus();
}

// Move the board to the position after the next move.
// Note that this does not only perform the move, but 
// also restores any other changes on the board.
// It does NOT change the annotations.
function nextMove()
{
    var dropdown = document.getElementById('movelist');
    if (dropdown.selectedIndex < dropdown.length-1) {
        dropdown.selectedIndex = dropdown.selectedIndex+1
        // Restore the position before next move...
        board.position(dropdown.options[dropdown.selectedIndex].value);
        // And perform the move
        checkAndDoMove(board,dropdown.options[dropdown.selectedIndex].text)
        updateStatus();
    }
}

// Move the board to the position after the last move.
// Note that this does not only perform the move, but 
// also restores any other changes on the board.
// It does NOT change the annotations.
function afterLastMove()
{
    var dropdown = document.getElementById('movelist');
    if (dropdown.length > 0) {
        // Select lasts move
        dropdown.selectedIndex = dropdown.length-1
        // Set the dummy board and do the last move on the dummyboard
        dummyboard.position(dropdown.options[dropdown.selectedIndex].value);
        checkAndDoMove(dummyboard,dropdown.options[dropdown.selectedIndex].text)
        // Copy position to the real board.
        board.position(dummyboard.position())
        updateStatus();
    }
}

// EVENTS HANDLERS

// General keys for the whole User Interface.
function doKeyPress(event)
{
    if (event.key === 'L') {  // Hide left pane
        document.getElementById('left_col').style.visibility = 'hidden'
    }
    if (event.key === 'l') {  // Show left pane
        document.getElementById('left_col').style.visibility = 'visible'
    }
    if (event.key === 'R') {  // Hide right pane
        document.getElementById('right_col').style.visibility = 'hidden'
    }
    if (event.key === 'r') {  // Show right pane
        document.getElementById('right_col').style.visibility = 'visible'
    }
    if (event.key === 'B') {  // Hide bottom right pane
        document.getElementById('bottom_right').style.visibility = 'hidden'
    }
    if (event.key === 'b') {  // Show bottom right pane
        document.getElementById('bottom_right').style.visibility = 'visible'
    }
    if (event.key === 's') {  // Start opstelling
        board.start()
    }
    if (event.key === 'e') {  // Leeg veld
        board.clear()
    }
    if (event.key === 'o') {  // Open file
        document.getElementById('loadfile').click()
    }
    if (event.key === 'p') {  // Vorige (previous) stelling
        previousPosition()
    }
    if (event.key === 'n') {  // Volgende (next) stelling
        nextPosition()
    }
    if (event.key === 'z') {   // Doe een zet
        nextMove()
    }
    if (event.key === 'Z') {  // Zet de laatste zet terug
        previousMove()
    }
    if (event.key === 'A') {  
        board.clearAnnotation()
    }
    if (event.key === 'f') {
        board.flip()
    }
    if (event.key === 't') {  // Zet de huidige stelling terug, ook de moves
        resetCurrentPosition()
    }
    if (event.key === 'T') {  // Zet de huidige stelling terug, met de huidige zet
        newMoveSelected()
    }
    if (event.key === 'h') {
        window.open('digidemobord-handleiding.html')
    }
}

// If 'Enter' is pressed in the FEN field, the board is updated.
function doKeyPressInputFEN(event)
{
    if (event.key === 'Enter') {
        fenToBoard()
    }
    event.stopPropagation();
}

// Searches for the first occurrence of a position with the current name...
// Returns true if found, selectedIndex will have the value of the 
// found occurence.
// Returns false if not found (selectedIndex is not updated).
// Search starts at currently selected position.
function searchPositionWithName(arg)
{
    var pos_name;
    if (arguments.length === 0)
        pos_name = document.getElementById('position_name').value
    else
        pos_name = arg

    var dropdown = document.getElementById('position_list');

    for (var i = dropdown.selectedIndex; i < dropdown.length; ++i) {
        // Skip case where there is no selection
        if (i < 0) 
            continue;
        if (dropdown.options[i].text === pos_name) {
            if (i !== dropdown.selectedIndex) {
                dropdown.selectedIndex = i;
                newPositionSelected();
            }
            return true;
        }
    }
    for (var i = 0; i < dropdown.selectedIndex; ++i) {
        if (dropdown.options[i].text === pos_name) {
            dropdown.selectedIndex = i;
            newPositionSelected();
            return true;
        }
    }
    return false;
}

// When 'Enter' is pressed in the position name, the following is done:
// * Search for the corresponding name.
// * If the name does not exist --> add current position
function doKeyPressPosition(event)
{

    if (event.key === 'Enter') {
        var name = document.getElementById('position_name').value;
        if (name !== '' && !searchPositionWithName()) {
            addPosition()
            document.getElementById('register_moves').checked  = true
            registerMovesStatusChanged()
        }
    }
    event.stopPropagation();
}

// When 'Enter' is pressed in the moves text input name, the moves
// of the current position are updated (replaced).
function doKeyPressMoves(event)
{
    if (event.key === 'Enter') {
        updateMoves();
    }
    event.stopPropagation();
}

// Adds the pieces (given as text) to the positions on the board.
function doAddPieces(side, pieces)
{
    // Replace any punctuation chars with spaces.
    pieces = pieces.replace(/\./g,' ')
    pieces = pieces.replace(/,/g,' ')
    pieces = pieces.replace(/;/g,' ')

    // Get board position
    position = board.position()
    // Get the pieces...
    pieces = pieces.split(' ');
    // For each piece
    for (var i = 0; i < pieces.length; ++i) {
        // Check the code
        var piece_string = pieces[i].trim()
        if (piece_string !== '') {
            // Get last two characters (i.e. the square)
            var square = piece_string.slice(-2);
            // If the first characters are a valid piece..
            if (validPieceUI(piece_string.slice(0,-2))) {
                // Create the piece code
                var piece = side + pieceUIToBoard[piece_string.slice(0,-2)];
                // If the square and piece are valid...
                if (validSquare(square) && validPieceCode(piece)) {
                    // Add to the position...
                    position[square] = piece;
                }
            }
        }
    }
    // Update the board based on the new position.
    board.position(position)
}

// If enter is pressed on the white pieces text inbox,
// add the pieces to the board.
function doKeyPressWhitePieces(event)
{
    if (event.key === 'Enter') {
        // Don't track moves when adding pieces ...
        document.getElementById('register_moves').checked  = false
        doAddPieces('w', document.getElementById('white_pieces').value);
    }
    event.stopPropagation();
}

// If enter is pressed on the black pieces text inbox,
// add the pieces to the board.
function doKeyPressBlackPieces(event)
{
    if (event.key === 'Enter') {
        // Don't track moves when adding pieces ...
        document.getElementById('register_moves').checked  = false
        doAddPieces('b', document.getElementById('black_pieces').value);
    }
    event.stopPropagation();
}

// Returns all pieces of a certain color (either 'w' or 'b') 
// that are currently on the board. These are returned in a text string, 
// to be added to the user interface.
//
// E.g. "a2 b2 c2 Dd1 Ke2 Th1"
function getPiecesAsString(color) {
    var position = board.position();
    var pieces_string = '';
    // For all squares...
    for (square in position) {
        var piece = position[square];
        // If right color
        if (piece[0] === color) {
            // Add the piece
            pieces_string += ' ' + pieceBoardToUI[piece[1]] + square;
        }
    }
    return pieces_string.trim();
}

// Prevents key propagations for the text input for pattern-matching
// based searching of position names...
function doKeyPressSearchPosition(event)
{
    event.stopPropagation();
}

// Keyup event for text input for pattern-matching based
// search of position names. Apparently, key-up is required
// to catch 'Escape' (does not work with keypress ??!!)
function doKeyUpSearchPosition(event)
{
    // Depending on the key ...
    switch (event.key) {
        // Enter means: select!
        // The currently selected (or first one) of the dropdown is used
        // If nothing is selected, and nothing in the dropdown, nothing is performed.
        case 'Enter':
            var dropdown = document.getElementById('match_list');
            var selected = dropdown.selectedIndex;
            if (selected < 0 && dropdown.length > 0)
                selected = 0
            if (selected > -1) {
                searchPositionWithName(dropdown.options[selected].text);
            }
            statusPositionReset()
            break;
        // Escape means: reset and ignore any selectoin
        case 'Escape':
            statusPositionReset()
            break;
        // Arrows can be used to move up and down the dropdown selection
        case 'ArrowUp':
            var matchdropdown = document.getElementById('match_list');
            matchdropdown.selectedIndex = matchdropdown.selectedIndex - 1;
            if (matchdropdown.selectedIndex < 0)
                matchdropdown.selectedIndex = 0;
            break;
        case 'ArrowDown':
            var matchdropdown = document.getElementById('match_list');
            if (matchdropdown.selectedIndex+1 < matchdropdown.length)
                matchdropdown.selectedIndex = matchdropdown.selectedIndex + 1;
            break;
        default:
            // In all other cases, 
            // update list with pattern matched results...
            var namesdropdown = document.getElementById('position_list');
            var pattern = document.getElementById('search_status_position').value
            clearList('match_list');
            for (var i = 0; i < namesdropdown.length; ++i)
            {
                var regexp = RegExp(pattern)
                if (namesdropdown.options[i].text.match(regexp))
                   addOptionToList('match_list', namesdropdown.options[i].text,"")
            }
    }
    event.stopPropagation();
}

// KEY EVENTS

// Global key events
const body_node = document.getElementById('body');
body_node.addEventListener('keypress', doKeyPress);

// Events listener (to ENter) for inputfen --> set board to fen
document.getElementById('inputfen').addEventListener('keypress', doKeyPressInputFEN)

// Events listener (to Enter) for stellingname --> adds stelling to list
document.getElementById('position_name').addEventListener('keypress', doKeyPressPosition)

// Events listener for moves --> prevents that any typing in this square leads
// to global event interpretation
document.getElementById('moves').addEventListener('keypress', doKeyPressMoves)

// Events listener for white and black pieces
// to global event interpretation
document.getElementById('white_pieces').addEventListener('keypress', doKeyPressWhitePieces)
document.getElementById('black_pieces').addEventListener('keypress', doKeyPressBlackPieces)

// keypress does not seem to work for Escape?!
document.getElementById('search_status_position').addEventListener('keyup', doKeyUpSearchPosition)
document.getElementById('match_list').addEventListener('keyup', doKeyUpSearchPosition)
// And keypress required to prevent passing the keypress to other elements..
document.getElementById('search_status_position').addEventListener('keypress', doKeyPressSearchPosition)

/*
 * Other functions ...
 */

// Updates the status line, containing position and last move
function updateStatus() {
    // Initialize info...
    var pos_name = ''
    var zet = ''

    // Look for info in the position_list..
    var dropdown = document.getElementById('position_list');
    if (dropdown.selectedIndex > -1) 
      pos_name = dropdown.options[dropdown.selectedIndex].text;

    // Look for info in the movelist...
    dropdown = document.getElementById('movelist');
    if (dropdown.selectedIndex > -1) {
      var move = dropdown.options[dropdown.selectedIndex].text;
      zet = ' (na ' + stripMoveDecoration(move) + ')';
    }

    // In case no move has been played, try to find out
    // who is to move next, and display that.
    var to_move = ''
    if (dropdown.selectedIndex < 0 && dropdown.options.length > 0) {
      var move = stripMoveDecoration(dropdown.options[0].text);
      to_move = whoMovesNext(board, move)
      if (to_move !== '' && validPlayer(to_move))
      {
         to_move = ' [' + playerToUI[to_move] + ']'
      }
    }

    // Update the status_position
    document.getElementById('status_position').innerHTML = 
        pos_name + zet + to_move
}

// Resets the current stelling in the list, and updates the board, 
// FEN, position_name and moves (list and field)
// It used the position_list for the information.
function resetCurrentPosition() {

    // Resetting a position may change the board. These changes
    // should NOT be registered, thus we uncheck register_moves first
    document.getElementById('register_moves').checked  = false
    registerMovesStatusChanged()

    // Look for the right position ...
    var dropdown = document.getElementById('position_list');
    // If nothing selected
    if (dropdown.selectedIndex < 0) {
        if (dropdown.options.length === 0) {
            // Reset values
            document.getElementById('inputfen').value = '';
            document.getElementById('position_name').value = '';
            document.getElementById('moves').value = '';
            document.getElementById('white_pieces').value = '';
            document.getElementById('black_pieces').value = '';
            return
        }
        // Select first one..
        dropdown.selectIndex = 0
    }

    // Get all information (fen, moves) from the position_list options
    var fenandmoves = dropdown.options[dropdown.selectedIndex].value;
    var fenstring = fenfromlistvalue(fenandmoves)
    var movesstring = movefromlistvalue(fenandmoves)
    var stelling = dropdown.options[dropdown.selectedIndex].text;
    document.getElementById('inputfen').value = fenstring
    document.getElementById('position_name').value = stelling
    document.getElementById('moves').value = movesstring

    // Reset the moves
    var dropdown = document.getElementById('movelist');
    dropdown.selectedIndex = -1;

    // Pas bord aan
    fenToBoard()

    // Voeg de stukken toe.
    document.getElementById('white_pieces').value = getPiecesAsString('w');
    document.getElementById('black_pieces').value = getPiecesAsString('b');

    // En info regel
    updateStatus();
}

// Updates the board (and other fields) after a new position
// has been selected. Uses resetCurrentPosition() for most
// of the work, and additionally clears the annotations
// and rebuilds the movelist.  
function newPositionSelected() {
    // Do all the work for resetting to the current position
    resetCurrentPosition()

    // remove annotation
    board.clearAnnotation()

    // Clear the movelist
    clearList('movelist');

    // Look for the right position ...
    var dropdown = document.getElementById('position_list');
    if (dropdown.selectedIndex < 0) 
        return;

    // Get information (fen, moves) from the position_list options
    var fenandmoves = dropdown.options[dropdown.selectedIndex].value;
    var fenstring = fenfromlistvalue(fenandmoves)
    var movesstring = movefromlistvalue(fenandmoves)

    // Update the movelist
    var moves = movesstring.split(' ');
    dummyboard.position(fenstring)
    for (var i = 0; i < moves.length; ++i) {
        var move = moves[i];
        // Position before move..
        var newfen = dummyboard.fen()
        // Perform the move
        move = checkAndDoMove(dummyboard, move);
        // move could be empty, if the move is not a legal move.
        if (move === '') continue 
        // Determine whether white or black played
        var white = whitePlayed(dummyboard, move)
        // Add to movelist
        addOptionToList('movelist', addMoveDecoration(move, white), newfen)
    }

    // Update status info
    updateStatus();

    // reset the color of the status line
    document.getElementById('status_position').style.color = "black"
}

// Restores the right position after the currently selected move.
// Does not really perform the move, as several things may have happened.
function newMoveSelected() {
    var dropdown = document.getElementById('movelist');
    if (dropdown.selectedIndex >= 0) {
        // Restore position after move
        if (dropdown.selectedIndex < dropdown.options.length-1)
            board.position(dropdown.options[dropdown.selectedIndex+1].value);
        else {
            // Last move: no position stored, so go to the position before
            // and exectute the move.
            dummyboard.position(dropdown.options[dropdown.selectedIndex].value);
            checkAndDoMove(dummyboard, dropdown.options[dropdown.selectedIndex].text);
            board.position(dummyboard.position())
        }
    } else {
        // Restore original position
        var dropdown = document.getElementById('position_list');
        if (dropdown.selectedIndex >= 0)
            board.position(fenfromlistvalue(dropdown.options[dropdown.selectedIndex].value))
    }
    updateStatus();
}

// Get the FEN from the stelling list value
function fenfromlistvalue(value) {
    if (value.split(' # ').length > 0) {
        return value.split(' # ')[0]
    } else {
        return '';
    }
}

// Get the moves from the stelling list value
function movefromlistvalue(value) {
    if (value.split(' # ').length > 1) {
        return value.split(' # ')[1]
    } else {
        return '';
    }
}

// Event function for opening a file.
function openFileEvent(event)
{
    // Get the files
    var files = event.target.files;
    // In case of cancel, files.length === 0
    if (files.length > 0)
        // Open them (only 1)
        openFile(files);
}

// Event function for reloading the current file.
function reloadFile()
{
    // Reloads the current file
    var loadfile = document.getElementById('loadfile')
    openFile(loadfile.files)
}

// Clears the position list, filename etc.
function clearFile()
{
    // Reset the current loaded file name
    var loadfile = document.getElementById('loadfile');
    loadfile.value = '';

    // Open those, which will reset position list, move list and status.
    openFile(loadfile.files)

    // Update status etc.
    updateMoveAndPositionButtonsVisibility()
    newPositionSelected();
    updateStatus();
}

// Does the parsing of the file contents.
function readV10File(lines)
{
    for (var i = 0; i < lines.length; ++i) {
        var line = lines[i].trim()
        // Only process non-empty lines that are not commented
        if ( (line != '') && (line[0] != '#') ) {
            var keyvalue = line.split(' | ');
            // Add the FEN and moves in one string. These will be interpreted by newPositionSelected()!
            if (keyvalue.length === 2)
                addOptionToList('position_list', keyvalue[0].trim(), keyvalue[1].trim());
        }
    }
}

// Reads positions and moves from a file.
// Current format:
//   <stelling-name> | <FEN> # <moves>
// on one line.
// Lines starting with a # are assumed to be comments, 
// the # and moves do not need to be present (moves are optional).
// Input files are text files, and can thus be edited manually.
// There is not much checking of the file format, and contents, so
// in case of errors in the file, you may have a problem.
function openFile(input) {
    // var input = event.target;
    
    clearList('position_list');
    clearList('movelist')
    // No file case ...
    if (input.length <= 0 || input[0] == '') {
        document.getElementById('filename').innerHTML = ''
        document.getElementById('status_filename').innerHTML = '[geen bestand]';
        return
    }

    var reader = new FileReader();
    reader.onload = function(){
        var text = reader.result;
        var lines = text.split('\n');
        // Test the first line for a version number
        if (lines[0].match('# ddm-1.0')) {
            // First version with a header
            lines.shift()
            readV10File(lines)
        } else {
            // Zero version: no header
            readV10File(lines)
        }
        // if positions read ...
        if (document.getElementById('position_list').options.length > 0) {
          // Start with first stelling after reading a file.
          document.getElementById('position_list').selectedIndex = 0
        }
        // moves and position buttons are only visible if there are positions
        updateMoveAndPositionButtonsVisibility()
        newPositionSelected()
    };
    // Things that are put here may be executed *before* the reader has finished (?!)
    reader.readAsText(input[0]);
    document.getElementById('filename').innerHTML = input[0].name;
    document.getElementById('status_filename').innerHTML = input[0].name + ':&nbsp';
    updateStatus()
};


// For downloading a file...
// From the internet ...
function download(filename, text) {
    var element = document.createElement('a');
    element.setAttribute('href', 'data:text/plain;charset=utf-8,' + encodeURIComponent(text));
    element.setAttribute('download', filename);
    
    element.style.display = 'none';
    document.body.appendChild(element);
    
    element.click();
    
    document.body.removeChild(element);
} 

// Produces the text that represents the current list of stellingen
// and allows the user to save this text to a file via a download
// (it is not possible to directly save to a local drive via JavaScript).
function getFENFileText() {
    var dropdown = document.getElementById('position_list');
    text = '# ddb-1.0: version 1.0 of digidemobord files\n';
    for (var i = 0; i < dropdown.options.length; ++i)  {
        option = dropdown.options[i]
        text += option.text + ' | ' + option.value + '\n';
    }
    document.getElementById('fenfile').value = text;
    document.getElementById('fenfile').select();
    
    /* Copy the text inside the text field */
    document.execCommand('copy');

    // Allow user to download the file. 
    download('stellingen.txt', text)
};

// Removes all options from a list.
function clearList(list) {
    var select = document.getElementById(list);
    select.options.length = 0
};

// Adds an option to a list
function addOptionToList(list, text, value) {
    
    // Skip empty labels
    if (text.length <= 0) {
        return
    }
    
    // get reference to select element
    var sel = document.getElementById(list);
    // create new option element
    var opt = document.createElement('option');
    // create text node to add to option element (opt)
    opt.appendChild( document.createTextNode(text) );
    // set value property of opt
    opt.value = value; 
    // add opt to end of select box (sel)
    sel.appendChild(opt);
};

// Adds some (none-alpha) chars to a move, to make the
// display in the dropdown list better (e.g. to make 
// black moves indented).
function addMoveDecoration(move, white)
{
    var prefix = white ? '> ' : '-------> '
    return prefix+move
}

// Strips the (none-alpha) chars from a move. 
// Moves in the movelist have decoration, to make the
// moves of black appear to the right (see also
// addMoveDecoration()).
// This function strips that info, to get a normal
// string representation of the move.
function stripMoveDecoration(move)
{
    // Remove any prefix garbage, added because of the dropdown list representation
    // All non-chars are removed.
    var start = move.search(/[a-zA-Z]/)
    if (start >= 0)
        move = move.slice(start,move.length)
    // Remove any remaining space...
    return move.trim()
}

// There are various textual representations of moves in this
// application:
//  * g1-f3 : simple (standard) representation, that is accepted
//            by a board
//  * Pf3xd5 : advanced representation (including piece, and a x
//             it is move that takes an opponents piece. This also
//             includes:
//              * o-o : castling, and similarly promotion
//              * e7-e8P : (minor) promotion)
//              * e5xd6,ep : en passent
//  * decorated style, which is the above, with some stuff prepended
//    for display in the movelist. 
// This function converts any move from the above format to a simple move
// that can be done by the Board (which are the start and end square
// separated by a '-', and no additional info). 
// Any other characters (non-alpha) that are in front of the
// move will be removed. These may be in front in case the move
// is shown in the dropdown list.
// And the piece annotation at the start and any info at the back 
// (except for minor promotion) will be stripped. 
// Castling notations will be replaced by a move by the King (e1-g1).
// Also a cross will be replaced by a dash.
function getStandardMove(move) 
{
    move = stripMoveDecoration(move);

    // special cases: castling
    if (move === 'o-o') return 'e8-g8';
    if (move === 'o-o-o') return 'e8-c8';
    if (move === 'O-O') return 'e1-g1';
    if (move === 'O-O-O') return 'e1-gc';
    
    // Split at '-' or 'x'
    var squares = move.split('-')
    if (squares.length < 2)
        squares = move.split('x')

    if (squares.length !== 2)
        return ''

    if (squares[0].length < 2 || squares[0].length > 3)
        return ''
    if (squares[1].length < 2 || squares[0].length > 5)
        return ''

    // Combine both squares (stripping prefix or postfix)
    var start = squares[0].slice(-2)
    var end = squares[1].slice(0,2)
    if (!validSquare(start) || !validSquare(end))
        return ''

    // Not a valid piece code ...
    if (squares[0].length === 3 && !validPieceUI(squares[0][0]))
        return ''

    // Rest could still be wrong, but is silently ignored.....
    var newMove = squares[0].slice(-2)+'-'+squares[1].slice(0,2)

    // Assume a (minor-)promotion, keep the piece...
    if (squares[1].length === 3 && validPieceUI(squares[1][2])) 
       newMove += squares[1][2]

    return newMove
}

// Checks whether the previous move was done by white 
// (if the ending square of the move is a white piece).
// Invalid moves or empty start squares also return true.
function whitePlayed(board, move)
{
    var standardMove = getStandardMove(move)
    if (standardMove === '')
        return true
    var end = standardMove.split('-')[1]
    var position = board.position()
    // This should not happen, but could happen if 
    // the moves do not match with the position ...
    // As moves can be typed, this can changed.
    if (position[end] != undefined)
      return position[end][0] === 'w'

    return true
}

// Determines the player who will execute
// the next move, by checking the color of
// the piece that is at the starting square
function whoMovesNext(board, move)
{
    // Get a decent move representatin
    var standardMove = getStandardMove(move)
    if (standardMove === '')
        return ''
    // Get start square
    var start = standardMove.split('-')[0]
    var position = board.position()
    // This should not happen, but could happen if 
    // the moves do not match with the position ...
    // As moves can be typed, this can changed.
    if (position[start] != undefined) {
        return position[start][0];
    }

    return ''
}

// Copied from chessboard
function isString (s) {
    return typeof s === 'string'
}

// Copied from chessboard
function validSquare (square) {
    return isString(square) && square.search(/^[a-h][1-8]$/) !== -1
}

// Copied from chessboard
function validPieceCode (code) {
    return isString(code) && code.search(/^[bw][KQRNBP]$/) !== -1
}

// Conversion from column name to an index (number)
var col2num = {
    'a' : 1, 'b' : 2, 'c' : 3, 'd' : 4,
    'e' : 5, 'f' : 6, 'g' : 7, 'h' : 8
}

// Converts the character for a piece on the board (English) to a 
// character for the UI (which is Dutch)
var pieceBoardToUI = {
    'K' : 'K', 'Q' : 'D', 'B' : 'L', 'N' : 'P', 'R' : 'T', 'P' : ''
}

// Converts the character for a player on the board (English) to a 
// character for the UI (which is Dutch)
var playerToUI = {
    'w' : 'W', 'b' : 'Z'
}

// Converts the character for a piece in the UI (Dutch) to a 
// character for the board (which is English)
var pieceUIToBoard = {
    'K' : 'K', 'D' : 'Q', 'L' : 'B', 'P' : 'N', 'T' : 'R', '' : 'P'
}

// Is the piece a valid piece for the Board?
// This is done by checking the keys of pieceBoardToUI.
// Valid pieces are K, Q, B, N, P, R and P.
function validPieceBoard(piece) {
    for (var p in pieceBoardToUI)
        if (p === piece) 
            return true;
    return false;
}

// Checks whether the piece is a valid piece in the UI
// (by checking the keys of pieceUIToBoard)
function validPieceUI(piece) {
    for (var p in pieceUIToBoard)
        if (p === piece) 
            return true;
    return false;
}

// Returns  whether a player is valid (in Board terminology)
// (by checking the keys of playerToUI)
function validPlayer(player) {
    for (var p in playerToUI)
        if (p === player)
            return true;
    return false;
}

// Given a move (in any format, see above) and a board, the
// move will be done on the board, and a nice representation of the 
// move (see above) will be returned in text.
// 
// This function tries to be smart, i.e. King moves such as e1-e3 are 
// interpreted as castling, a promotion is automatically done (to Queen), 
// and in case of en-passant moves (e5-d6, with a black pawn on d5) will
// also result in removing the black pawn.
// 
// This function is also used after reading a text file, and while parsing
// the text string. Therefore, it will also do some sanity-checks, and will
// skip / ignore any text that can not easily be interpreted as a move.
//
// To Do: split the checking of the move, and performing the move?!
function checkAndDoMove(board, general_move) {
    
    // For convenience, get the simple representation
    var move = getStandardMove(general_move)
    if (move === '')
        return ''

    // Get current board position
    position = board.position()

    // Get start and end square of the move
    var squares = move.split('-')
    var start = squares[0]
    var end = squares [1]

    // Check whether the move at least contains sufficient info 
    // for the squares
    if (squares[0].length < 2 || squares[1].length < 2)
        return ''

    var prefix = ''   // Symbol for piece, none for pawn
    var move_start = squares[0].slice(-2)
    var sep = '-'     // 'x' is a fiendly piece was taken
    var move_end = squares[1].slice(0,2)
    var postfix = ''  // (minor)promotion or ,ep

    // If the squares are not valid squares ...
    if (!validSquare(move_start) || !validSquare(move_end))
        return ''

    // At least the squares are ok, other info may/will be ignored
    // if it can not be interpreted.

    // Check for special moves.
    if (start === 'e1' && end === 'g1' && position['h1'] === 'wR') {
        // castling white short
        board.move(move);
        board.move('h1-f1');
        return 'O-O'
    }
    else if (start === 'e1' && end === 'c1' && position['a1'] === 'wR') {
        // castling white long
        board.move(move);
        board.move('a1-d1');
        return 'O-O-O'
    }
    else if (start === 'e8' && end === 'g8' && position['h8'] === 'bR') {
        // castling black short
        board.move(move);
        board.move('h8-f8');
        return 'o-o'
    }
    else if (start === 'e8' && end === 'c8' && position['a8'] === 'bR') {
        // castling black long
        board.move(move);
        board.move('a8-d8');
        return 'o-o-o'
    } else if ( Math.abs(col2num[start[0]] - col2num[end[0]]) <= 1 
                && start[1] == '7' && end[1] == '8' && position[start] === 'wP') {
        // white promotion 
        // strip the promotion symbol, if present
        board.move(start+'-'+end.slice(0,2));
        // Update the position, i.e. replace the pawn with the promoted piece
        position = board.position()
        if (end.length > 2 && validPieceUI(end[2])) {  
          // new valid piece is given, use it
          position[end.slice(0,2)] = 'w' + pieceUIToBoard[end[2]];
          postfix = end[2];
        } else  {              
          // no new (or valid) piece is given, assume Queen
          position[end.slice(0,2)] = 'wQ';
          postfix = pieceBoardToUI['Q'];
        }
        if (start[0] != end[0]) sep = 'x';   // if pawn changed column, we need a cross
        board.position(position)
    } else if ( Math.abs(col2num[start[0]] - col2num[end[0]]) <= 1 
                && start[1] == '2' && end[1] == '1' && position[start] === 'bP') {
        // black promotion 
        // strip the promotion symbol, if present
        board.move(start+'-'+end.slice(0,2));
        // Update the position, i.e. replace the pawn with the promoted piece
        position = board.position()
        if (end.length > 2 && validPieceUI(end[2])) {      
          // new and valid piece is given, use it
          position[end.slice(0,2)] = 'b' + pieceUIToBoard[end[2]]
          postfix = end[2];
        } else {                   
          // no new (or valid) piece is given, assume Queen
          position[end.slice(0,2)] = 'bQ'
          postfix = pieceBoardToUI['Q'];
        }
        if (start[0] != end[0]) sep = 'x';  // column change implies 'x'
        board.position(position)
    } else if (start[1] === '5' && end[1] === '6' &&       // from 5th to 6th row
               ( Math.abs(col2num[start[0]] - col2num[end[0]]) === 1 ) &&  // moving diagonal
               position[start] === 'wP' &&                    // it is a pawn
               position[end[0]+'5'] === 'bP' &&               // there is a fiendly pawn at the right loc
               (typeof position[end] === 'undefined' || position[end] === '' ) )  // end pos is empty
    {
        // En passant for white
        // Move the pawn
        board.move(move);
        // Remove the other pawn ...
        position = board.position()
        delete position[end[0]+'5']
        board.position(position)
        sep = 'x';
        postfix = ',ep';
    } else if (start[1] === '4' && end[1] === '3' &&       // from 5th to 6th row
               ( Math.abs(col2num[start[0]] - col2num[end[0]]) === 1 ) &&  // moving diagonal
               position[start] === 'bP' &&                    // it is a pawn
               position[end[0]+'4'] === 'wP' &&               // there is a fiendly pawn at the right loc
               (typeof position[end] === 'undefined' || position[end] === '' ) )  // end pos is empty
    {
        // En passant for black
        // Move the pawn
        board.move(move);
        // Remove the other pawn ...
        position = board.position()
        delete position[end[0]+'4']
        board.position(position)
        sep = 'x';
        postfix = ',ep';
    }
    else {
        // Normal case
        position = board.position()
        // Get piece that is moved
        if ( position[move_start] != undefined ) {
            prefix = pieceBoardToUI[position[move_start][1]]
        } 
        // Check whether we need a cross
        if ( position[move_end] != undefined && position[move_end] != '')
            sep = 'x'
        // Perform move
        board.move(move);
    }
    // Return the move in nice format:
    return prefix + move_start + sep + move_end + postfix
}
